#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <exec/exec_protos.h>
#include <exec/io.h>
#include <dos/dos.h>

LONG _PrintFmt(va_list ap, void *fp, int (*cf)(int, void *), const char *fmt);
LONG _FPrintF(APTR fp, const char *fmt, ...);
LONG _PrintF(const char *fmt, ...);
struct ArgList *_ParseArgs(const char *);
void _FreeArgs(struct ArgList *);

#define DOSL_NUMVECS 61
#define DOSL_NEGSIZE (DOSL_NUMVECS*LIB_VECTSIZE)
#define DOSL_VERSION 1

#define HANDLER_SIZE 32 /* MaxSize of the handler name */

struct DOSLibrary *DOSOpen(void);
void DOSClose(void);
void InitDosLibrary(void);

static int HandlerPart(char *dest, const char *src, int size);
static char *FilePart(const char *src);

static int HandlerPart(char *dest, const char *src, int size)
{
    const char *a = src;
    int csize;

    while(*a)
    {
	if(*a==':')
	{
	    csize = (int)(a-src+1);
	    if(csize >= size)
	    {
		csize = size-1;
	    }
	    memcpy(dest, src, csize);
	    dest[csize] = '\0';
	    strtoupper(dest, dest);
	    return -1;
	}
	a++;
    }
    return 0;
}

static char *FilePart(const char *src)
{
    const char *a = src;

    while(*a)
    {
	if(*a==':')
	{
	    return (char *)a+1;
	}
	a++;
    }
    return (char *)src;
}


/* The library base is global because C can't easily get register params */
struct DOSLibrary *DOSBase = NULL;

void InitDosLibrary()
{
    char *base = (char *)AllocMem(sizeof(struct DOSLibrary)+DOSL_NEGSIZE, MEMF_CLEAR);

    static const ULONG vectors[DOSL_NUMVECS] =
    {
	(ULONG)DOSOpen,
	(ULONG)DOSClose,
	(ULONG)DeviceProc,
	(ULONG)DoPkt4,
	(ULONG)DoPkt4Safe,
	(ULONG)Open,
	(ULONG)Close,
	(ULONG)Read,
	(ULONG)Write,
	(ULONG)Seek,
	(ULONG)SetMode,
	(ULONG)Input,
	(ULONG)Output,
	(ULONG)Error,
	(ULONG)Lock,
	(ULONG)UnLock,
	(ULONG)DupLock,
	(ULONG)SameLock,
	(ULONG)Examine,
	(ULONG)ExNext,
	(ULONG)ExamineEnd,
	(ULONG)CurrentDir,
	(ULONG)ParentDir,
	(ULONG)DeleteFile,
	(ULONG)Rename,
	(ULONG)SetComment,
	(ULONG)SetProtection,
	(ULONG)SetOwner,
	(ULONG)IoErr,
	(ULONG)SetIoErr,
	(ULONG)ROMLoadSeg,
	(ULONG)LoadSeg,
	(ULONG)UnLoadSeg,
	(ULONG)RunCommand,
	(ULONG)CreateProc,
	(ULONG)Delay,
	(ULONG)WaitForChar,
	(ULONG)DeleteVar,
	(ULONG)FindVar,
	(ULONG)GetVar,
	(ULONG)SetVar,
	(ULONG)CreateDir,
	(ULONG)PrintFault,
	(ULONG)Format,
	(ULONG)DiskInfo,
	(ULONG)SetFileDate,
	(ULONG)IsFileSystem,
	(ULONG)IsInteractive,
	(ULONG)Relabel,
	(ULONG)CreateNewProcTags,
	(ULONG)SystemTags,
	(ULONG)GetDeviceProc,
	(ULONG)FreeDeviceProc,
	(ULONG)AssignLock,
	(ULONG)AssignLate,
	(ULONG)AssignAdd,
	(ULONG)_PrintFmt,
	(ULONG)_FPrintF,
	(ULONG)_PrintF,
	(ULONG)_ParseArgs,
	(ULONG)_FreeArgs
    };

    if(base)
    {
	DOSBase = (struct DOSLibrary *)(base+DOSL_NEGSIZE);

	DOSBase->dl_lib.lib_Node.ln_Name = DOSL_NAME;
	DOSBase->dl_lib.lib_Node.ln_Pri = 0;
	DOSBase->dl_lib.lib_Node.ln_Type = NT_LIBRARY;
	DOSBase->dl_lib.lib_Version = DOSL_VERSION;
	DOSBase->dl_lib.lib_OpenCnt = 0;
	DOSBase->dl_Flags = 0;
	NewList(&DOSBase->dl_Handlers);
	InitLibFuncs(&DOSBase->dl_lib, (ULONG *)vectors, DOSL_NUMVECS);
	AddLibrary(&DOSBase->dl_lib);
    }
}

struct DOSLibrary *DOSOpen()
{
    Forbid();
    DOSBase->dl_lib.lib_OpenCnt++;
    Permit();
    return DOSBase;
}

void DOSClose(void)
{
    Forbid();
    DOSBase->dl_lib.lib_OpenCnt--;
    Permit();
}

static LONG DoPkt4(struct MsgPort *port,
		   LONG type,
		   ULONG arg1,
		   ULONG arg2,
		   ULONG arg3,
		   ULONG arg4)
{
    struct Process *proc = (struct Process *)FindTask(NULL);
    struct DosPacket dp;
    
    /* Get The packet ready */
    dp.dp_Type = type;
    dp.dp_Arg1 = arg1;
    dp.dp_Arg2 = arg2;
    dp.dp_Arg3 = arg3;
    dp.dp_Arg4 = arg4;
    dp.dp_Message.mn_Length = sizeof(struct DosPacket);
    dp.dp_Message.mn_ReplyPort = &proc->pr_Task.tc_Port;
    PutMsg(port, &dp.dp_Message);
    Wait((1<<(proc->pr_Task.tc_Port.mp_SigBit)));
    while(dp.dp_Message.mn_Node.ln_Type != NT_REPLYMSG)
	Wait((1<<(proc->pr_Task.tc_Port.mp_SigBit)));
    Remove((struct Node *)&dp.dp_Message);
    SetIoErr(dp.dp_Result2);
    return dp.dp_Result;
}

static LONG DoPkt4Safe(const char *handler,
		   LONG type,
		   ULONG arg1,
		   ULONG arg2,
		   ULONG arg3,
		   ULONG arg4)
{
    struct Process *proc = (struct Process *)FindTask(NULL);
    struct DosPacket dp;
    struct MsgPort *port;

    Forbid();
    if(!(port = DeviceProc(handler)))
    {
	Permit();
	proc->pr_Result2 = ERROR_OBJECT_NOT_FOUND;
	return NULL;
    }

    /* Get The packet ready */
    dp.dp_Type = type;
    dp.dp_Arg1 = arg1;
    dp.dp_Arg2 = arg2;
    dp.dp_Arg3 = arg3;
    dp.dp_Arg4 = arg4;
    dp.dp_Message.mn_Length = sizeof(struct DosPacket);
    dp.dp_Message.mn_ReplyPort = &proc->pr_Task.tc_Port;
    PutMsg(port, &dp.dp_Message);
    Permit();
    Wait((1<<(proc->pr_Task.tc_Port.mp_SigBit)));
    while(dp.dp_Message.mn_Node.ln_Type != NT_REPLYMSG)
	Wait((1<<(proc->pr_Task.tc_Port.mp_SigBit)));
    Remove((struct Node *)&dp.dp_Message);
    SetIoErr(dp.dp_Result2);
    return dp.dp_Result;
}

/*
   Create/Override/Delete an assign. If lock==NULL, deletes the assign.
   If there is an entry with the same name and it is NOT an assign,
   an error is returned.
   */

static LONG AssignLock(const char *assign, APTR lock)
{
    struct DosList *dl;
    struct AssignList *al;

    /* TODO: add a little more thrular check */
    if(assign[strlen(assign)-1]!=':')
    {
	SetIoErr(ERROR_INVALID_COMPONENT_NAME);
	return DOS_FALSE;
    }

    /* TODO: doslist locking */
    Forbid();
    dl = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dl->dol_Node.ln_Next)
    {
	if(!strcasecmp(dl->dol_Node.ln_Name, assign))
	{
	    if(dl->dol_Node.ln_Type == DLT_ASSIGN)
	    {
		if(dl->dol_misc.dol_assign.dol_AssignName)
		{
		    /* Late Binding cancelled !*/
		    FreeMem(dl->dol_misc.dol_assign.dol_AssignName,
			    strlen(dl->dol_misc.dol_assign.dol_AssignName)+1);
		    dl->dol_misc.dol_assign.dol_AssignName = NULL;
		}
		while((al = (struct AssignList *)
		       RemHead(&dl->dol_misc.dol_assign.dol_AssignList)))
		{
		    UnLock(al->al_Lock);
		    FreeMem(al, sizeof(struct AssignList));
		}
		if(!lock)
		{
		    Remove((struct Node *)dl);
		    FreeMem(dl->dol_Node.ln_Name,
			    strlen(dl->dol_Node.ln_Name)+1);
		    FreeMem(dl, sizeof(struct DosList));

		    /* TODO: doslist unlocking */
		    Permit();
		    return DOS_TRUE;
		}

		al = (struct AssignList *)
		    AllocMem(sizeof(struct AssignList), MEMF_CLEAR);
		if(!al)
		{
		    Remove((struct Node *)dl);
		    FreeMem(dl->dol_Node.ln_Name,
			    strlen(dl->dol_Node.ln_Name)+1);
		    FreeMem(dl, sizeof(struct DosList));

		    /* TODO: doslist unlocking */
		    Permit();

		    SetIoErr(ERROR_NO_FREE_STORE);
		    return DOS_FALSE;
		}
		al->al_Lock = lock;
		AddTail(&dl->dol_misc.dol_assign.dol_AssignList,
			(struct Node *)al);

		/* TODO: doslist unlocking */
		Permit();
		return DOS_TRUE;
	    }
	    else
	    {
		/* TODO: doslist unlocking */
		Permit();
		SetIoErr(ERROR_OBJECT_EXISTS);
		return DOS_FALSE;
	    }
	}
	dl = (struct DosList *)dl->dol_Node.ln_Next;
    }
    /* OK, there is not an entry with that name.
       If lock==NULL, this is an error */

    if(!lock)
    {
	/* TODO: doslist unlocking */
	Permit();

	SetIoErr(ERROR_OBJECT_NOT_FOUND);
	return DOS_FALSE;
    }

    if((dl = AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
    {
	if((al = AllocMem(sizeof(struct AssignList), MEMF_CLEAR)))
	{
	    if((dl->dol_Node.ln_Name = AllocMem(strlen(assign)+1, 0)))
	    {
		strcpy(dl->dol_Node.ln_Name, assign); /* Was strtoupper(); */
		dl->dol_Node.ln_Type = DLT_ASSIGN;
		NewList(&dl->dol_misc.dol_assign.dol_AssignList);
		/* dol_AssignName = NULL, dol_Flags = 0 */
		/* dol_Task and dol_Lock are not currently used for assigns */

		al->al_Lock = lock;
		AddTail(&dl->dol_misc.dol_assign.dol_AssignList,
			(struct Node *)al);

		AddTail(&DOSBase->dl_Handlers, (struct Node *)dl);

		/* TODO: doslist unlocking */
		Permit();
		return DOS_TRUE;
	    }
	    FreeMem(al, sizeof(struct AssignList));
	}
	FreeMem(dl, sizeof(struct DosList));
    }

    /* TODO: doslist unlocking */
    Permit();
    SetIoErr(ERROR_NO_FREE_STORE);
    return DOS_FALSE;
}

static LONG AssignLate(const char *assign, const char *path)
{
    struct DosList *dl;
    struct AssignList *al;

    /* TODO: add a little more thrular check */
    if(assign[strlen(assign)-1]!=':')
    {
	SetIoErr(ERROR_INVALID_COMPONENT_NAME);
	return DOS_FALSE;
    }

    /* TODO: doslist locking */
    Forbid();
    dl = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dl->dol_Node.ln_Next)
    {
	if(!strcasecmp(dl->dol_Node.ln_Name, assign))
	{
	    if(dl->dol_Node.ln_Type == DLT_ASSIGN)
	    {
		if(dl->dol_misc.dol_assign.dol_AssignName)
		{
		    /* Late Binding cancelled !*/
		    FreeMem(dl->dol_misc.dol_assign.dol_AssignName,
			    strlen(dl->dol_misc.dol_assign.dol_AssignName)+1);
		    dl->dol_misc.dol_assign.dol_AssignName = NULL;
		}
		while((al = (struct AssignList *)
		       RemHead(&dl->dol_misc.dol_assign.dol_AssignList)))
		{
		    UnLock(al->al_Lock);
		    FreeMem(al, sizeof(struct AssignList));
		}

		if((dl->dol_misc.dol_assign.dol_AssignName =
		    (char *)AllocMem(strlen(path)+1, 0)))
		{
		    strcpy(dl->dol_misc.dol_assign.dol_AssignName,
			   path);
		    
		    AddTail(&DOSBase->dl_Handlers, (struct Node *)dl);
		    /* TODO: doslist unlocking */
		    Permit();
		    return DOS_TRUE;
		}
		
		/* TODO: doslist unlocking */
		Permit();
		SetIoErr(ERROR_NO_FREE_STORE);
		return DOS_FALSE;
	    }
	    else
	    {
		/* TODO: doslist unlocking */
		Permit();
		SetIoErr(ERROR_OBJECT_EXISTS);
		return DOS_FALSE;
	    }
	}
	dl = (struct DosList *)dl->dol_Node.ln_Next;
    }
    /* OK, there is not an entry with that name. */

    if((dl = AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
    {
	if((dl->dol_Node.ln_Name = AllocMem(strlen(assign)+1, 0)))
	{
	    strcpy(dl->dol_Node.ln_Name, assign); /* Was strtoupper(); */
	    dl->dol_Node.ln_Type = DLT_ASSIGN;
	    NewList(&dl->dol_misc.dol_assign.dol_AssignList);
	    /* dol_AssignName = path, dol_Flags = 0 */
	    /* dol_Task and dol_Lock are not currently used for assigns */
	    
	    if((dl->dol_misc.dol_assign.dol_AssignName =
		(char *)AllocMem(strlen(path)+1, 0)))
	    {
		strcpy(dl->dol_misc.dol_assign.dol_AssignName,
		       path);

		AddTail(&DOSBase->dl_Handlers, (struct Node *)dl);
		/* TODO: doslist unlocking */
		Permit();
		return DOS_TRUE;
	    }
	    
	    /* TODO: doslist unlocking */
	    Permit();
	    SetIoErr(ERROR_NO_FREE_STORE);
	    return DOS_FALSE;
	}
	FreeMem(dl, sizeof(struct DosList));
    }

    /* TODO: doslist unlocking */
    Permit();
    SetIoErr(ERROR_NO_FREE_STORE);
    return DOS_FALSE;
}

static LONG AssignAdd(const char *assign, APTR lock)
{
    struct DosList *dl;
    struct AssignList *al;
    APTR lock2;

    if(assign[strlen(assign)-1]!=':')
    {
	SetIoErr(ERROR_INVALID_COMPONENT_NAME);
	return DOS_FALSE;
    }

    /* TODO: doslist locking */
    Forbid();
    dl = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dl->dol_Node.ln_Next)
    {
	if(!strcasecmp(dl->dol_Node.ln_Name, assign))
	{
	    if(dl->dol_Node.ln_Type == DLT_ASSIGN)
	    {
		if(dl->dol_misc.dol_assign.dol_AssignName)
		{
		    /* Late Binding !*/
		    lock2 = Lock(dl->dol_misc.dol_assign.dol_AssignName,
				 ACCESS_READ);

		    if(lock2)
		    {
			al = (struct AssignList *)
			    AllocMem(sizeof(struct AssignList), MEMF_CLEAR);
			if(al)
			{
			    al->al_Lock = lock2;
			    AddTail(&dl->dol_misc.dol_assign.dol_AssignList,
				    (struct Node *)al);
			    FreeMem(dl->dol_misc.dol_assign.dol_AssignName,
				    strlen(dl->dol_misc.dol_assign.dol_AssignName)+1);
			    dl->dol_misc.dol_assign.dol_AssignName = NULL;
			}
			else
			{
			    UnLock(lock2);
			    /* TODO: doslist unlocking */
			    Permit();
			    SetIoErr(ERROR_NO_FREE_STORE);
			    return DOS_FALSE;
			}
		    }
		    else
		    {
			/* TODO: doslist unlocking */
			Permit();
			SetIoErr(ERROR_OBJECT_NOT_FOUND);
			return DOS_FALSE;
		    }
		}

		al = (struct AssignList *)
		    AllocMem(sizeof(struct AssignList), MEMF_CLEAR);

		if(!al)
		{
		    /* TODO: doslist unlocking */
		    Permit();

		    SetIoErr(ERROR_NO_FREE_STORE);
		    return DOS_FALSE;
		}
		al->al_Lock = lock;
		AddTail(&dl->dol_misc.dol_assign.dol_AssignList,
			(struct Node *)al);

		/* TODO: doslist unlocking */
		Permit();
		return DOS_TRUE;
	    }
	    else
	    {
		/* TODO: doslist unlocking */
		Permit();
		SetIoErr(ERROR_OBJECT_EXISTS);
		return DOS_FALSE;
	    }
	}
	dl = (struct DosList *)dl->dol_Node.ln_Next;
    }
    /* OK, there is not an entry with that name. This could be an error */

    if((dl = AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
    {
	if((al = AllocMem(sizeof(struct AssignList), MEMF_CLEAR)))
	{
	    if((dl->dol_Node.ln_Name = AllocMem(strlen(assign)+1, 0)))
	    {
		strcpy(dl->dol_Node.ln_Name, assign); /* Was strtoupper(); */
		dl->dol_Node.ln_Type = DLT_ASSIGN;
		NewList(&dl->dol_misc.dol_assign.dol_AssignList);
		/* dol_AssignName = NULL, dol_Flags = 0 */
		/* dol_Task and dol_Lock are not currently used for assigns */

		al->al_Lock = lock;
		AddTail(&dl->dol_misc.dol_assign.dol_AssignList,
			(struct Node *)al);

		AddTail(&DOSBase->dl_Handlers, (struct Node *)dl);

		/* TODO: doslist unlocking */
		Permit();
		return DOS_TRUE;
	    }
	    FreeMem(al, sizeof(struct AssignList));
	}
	FreeMem(dl, sizeof(struct DosList));
    }

    /* TODO: doslist unlocking */
    Permit();
    SetIoErr(ERROR_NO_FREE_STORE);
    return DOS_FALSE;
}


static struct DevProc *GetDeviceProc(const char *filename, struct DevProc *old)
{
    struct DevProc *dev = old;
    struct Process *proc = (struct Process *)FindTask(NULL);
    struct DosList *dl;
    struct AssignList *al;
    char handler[HANDLER_SIZE];

    if(!dev)
    {
	dev = (struct DevProc *)AllocMem(sizeof(struct DevProc), MEMF_CLEAR);
    }
    if(!dev)
    {
	SetIoErr(ERROR_NO_FREE_STORE);
	return NULL;
    }

    if(!HandlerPart(handler, filename, HANDLER_SIZE))
    {
	/* relative path, use pr_FileSystemTask  "dir/file" */
	if(proc->pr_FileSystemTask)
	{
	    dev->dvp_Port = proc->pr_FileSystemTask;
	    dev->dvp_Lock = (ULONG)proc->pr_CurrentDir;
	    dev->dvp_Flags = 0;
	    dev->dvp_DevNode = 0;
	    /* TODO: find the node from doslist, so that multiassigns work.. */

	    return dev;
	}
    }
    if(!handler[1])
    {
	/* Relative to current handler  ":dir/file" */
	if(proc->pr_FileSystemTask)
	{
	    dev->dvp_Port = proc->pr_FileSystemTask;
	    dev->dvp_Lock = NULL;
	    dev->dvp_Flags = 0;
	    dev->dvp_DevNode = 0;
	    /* TODO: find the node from doslist, so that multiassigns work.. */

	    return dev;
	}
    }
    /*
       Otherwise it is a full path  ("handler:dir/file")
       and we have to search the doslist for it.
       */

    /* TODO: doslist locking */
    Forbid();
    if(dev->dvp_DevNode)
	dl = dev->dvp_DevNode;
    else
	dl = (struct DosList *)DOSBase->dl_Handlers.lh_Head;
    while(dl->dol_Node.ln_Next)
    {
	if(!strcasecmp(dl->dol_Node.ln_Name, handler))
	{
	    if(dl->dol_Node.ln_Type == DLT_ASSIGN)
	    {
		if(dl->dol_misc.dol_assign.dol_AssignName)
		{
		    /* Late Binding !*/
		    APTR lock2 = Lock(dl->dol_misc.dol_assign.dol_AssignName,
				      ACCESS_READ);
		    if(lock2)
		    {
			al = (struct AssignList *)
			    AllocMem(sizeof(struct AssignList), MEMF_CLEAR);
			if(al)
			{
			    al->al_Lock = lock2;
			    AddTail(&dl->dol_misc.dol_assign.dol_AssignList,
				    (struct Node *)al);
			    FreeMem(dl->dol_misc.dol_assign.dol_AssignName,
				    strlen(dl->dol_misc.dol_assign.dol_AssignName)+1);
			    dl->dol_misc.dol_assign.dol_AssignName = NULL;
			}
			else
			{
			    UnLock(lock2);
			}
		    }
		}
		al = (struct AssignList *)dl->dol_misc.dol_assign.dol_AssignList.lh_Head;

		if(dl==dev->dvp_DevNode && dev->dvp_Lock)
		{
		    while(al->al_Node.lmn_Next && al->al_Lock!=dev->dvp_Lock)
			al = (struct AssignList *)al->al_Node.lmn_Next;
		    if(al->al_Node.lmn_Next && al->al_Lock==dev->dvp_Lock)
			al = (struct AssignList *)al->al_Node.lmn_Next;
		}
		if(al->al_Node.lmn_Next) /* valid lock node */
		{
		    dev->dvp_Port = ((struct FileLock *)al->al_Lock)->fl_Task;
		    dev->dvp_Lock = al->al_Lock;
		    dev->dvp_Flags = 0;
		    dev->dvp_DevNode = dl;
		    /* TODO: doslist unlocking */
		    Permit();
		    return dev;
		}
		SetIoErr(ERROR_NO_MORE_ENTRIES);
		/* Currently checks for another node with the same name */
	    }
	    else
	    {
		/* If there is no task currently handling this msgport */
		if(!dl->dol_Task)
		{
		    if(dl->dol_Node.ln_Type==DLT_HANDLER)
		    {
			if(!dl->dol_misc.dol_handler.dol_SegList &&
			   dl->dol_misc.dol_handler.dol_Handler)
			{
			    dl->dol_misc.dol_handler.dol_SegList =
				LoadSeg(dl->dol_misc.dol_handler.dol_Handler);
			}
			/* If seglist ready, start the task.. */
			if(dl->dol_misc.dol_handler.dol_SegList)
			{
			    CreateProc(dl->dol_Node.ln_Name,
				       dl->dol_misc.dol_handler.dol_Priority,
				       dl->dol_misc.dol_handler.dol_SegList,
				       dl->dol_misc.dol_handler.dol_Stacksize,
				       (char *)NULL);
			    Delay(5);
			}
		    }
		}
		
		if(dl->dol_Task)
		{
		    dev->dvp_Port = dl->dol_Task;
		    dev->dvp_Lock = dl->dol_Lock;
		    /* dol_Lock may be NULL, but the handler should handle it.. */
		    dev->dvp_Flags = 0;
		    dev->dvp_DevNode = dl;
		    /* TODO: doslist unlocking */
		    Permit();
		    return dev;
		}
		Permit();
		FreeMem(dev, sizeof(struct DevProc));
		SetIoErr(ERROR_OBJECT_NOT_FOUND);
		return NULL;
	    }
	}
	dl = (struct DosList *)dl->dol_Node.ln_Next;
    }
    /* TODO: doslist unlocking */
    Permit();
    FreeMem(dev, sizeof(struct DevProc));
    SetIoErr(ERROR_OBJECT_NOT_FOUND);
    return NULL;
}

/*
   FreeDeviceProc(GetDeviceProc("name:", NULL)) is always safe,
   because the DevProc structure is either returned or deallocated
   by GetDeviceProc(), in which case NULL is returned and
   FreeDeviceProc(NULL) is safe.
 */
static void FreeDeviceProc(struct DevProc *dev)
{
    if(dev)
    {
	FreeMem(dev, sizeof(struct DevProc));
    }
}

static struct MsgPort *DeviceProc(const char *handler)
{
    struct DevProc *dev = GetDeviceProc(handler, NULL);
    struct MsgPort *port = NULL;

    if(dev)
    {
	port = dev->dvp_Port;
	FreeDeviceProc(dev);
    }
    return port;
}


/*
   Open() determines the right handler process and sends an open
   request to it. If the filename does not have an absolute path,
   the open request is sent to the handler currently in the
   pr_FileSystemTask field. The handler itself should use the
   pr_CurrentDir field to determine the starting point for the
   relative path. If the handler does not implement directories,
   pr_CurrentDir is not needed. However, to be able to "cd" to
   the handler, Lock() must still be implemented.

   The handler will get a DosPacket to its message port with a FileHandle
   structure attached. The handler should initialize the FileHandle
   properly.
*/

static APTR Open(const char *filename, ULONG mode)
{
    struct DevProc *dev;
    APTR handle = NULL;

    if((dev = GetDeviceProc(filename, NULL)))
    {
	handle = (APTR)DoPkt4(dev->dvp_Port, ACTION_OPEN,
			      (ULONG)dev->dvp_Lock, (ULONG)filename,
			      mode, 0);
	FreeDeviceProc(dev);
    }
    return handle;
}

/*
   Close() closes the associated filehandle. The handler should
   deallocate all structures allocated during the associated Open()
   call.
*/
static LONG  Close(APTR handle)
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    return DoPkt4(fh->fh_Port, ACTION_CLOSE, (ULONG)fh, 0, 0, 0);
}

/*
   Read() gets data from the filehandle. The number of characters
   actually read is returned or a negative value for fatal error.
   IoErr() may be used to determine the cause. If the handle is
   for an interactive stream, less characters than "size" may
   be returned. Return value 0 is interpreted as EOF.
 */
static LONG  Read(APTR handle, char *buffer, LONG size)
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    return DoPkt4(fh->fh_Port, ACTION_READ, (ULONG)fh,
		  (ULONG)buffer, (ULONG)size, 0);
}

/*
   Write() puts data to the filehandle. The number of characters
   actually written is returned or negative value is returned
   for fatal error. IoErr() may be used to determine the cause.
 */
static LONG  Write(APTR handle, const char *buffer, LONG size)
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    if(!fh)
	return size;
    return DoPkt4(fh->fh_Port, ACTION_WRITE, (ULONG)fh,
		  (ULONG)buffer, (ULONG)size, 0);
}

static LONG SetMode(APTR handle, LONG mode)
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    if(!fh)
	return DOS_FALSE;
    return DoPkt4(fh->fh_Port, ACTION_SCREEN_MODE, (ULONG)fh, mode, 0, 0);
}

/*
   Input() returns the current input stream of the process.
 */
static APTR  Input(void)
{
    return ((struct Process *)FindTask(NULL))->pr_CIS;
}

/*
   Output() returns the current output stream of the process.
 */
static APTR  Output(void)
{
    return ((struct Process *)FindTask(NULL))->pr_COS;
}


/*
   Error() returns the current error stream of the process.
 */
static APTR  Error(void)
{
    return ((struct Process *)FindTask(NULL))->pr_CES;
}

/*
   Seek() changes the read and write locations in a filesystem
   stream. Old position is returned in secondary result (IoErr()).
 */
static LONG Seek(APTR handle, LONG offset, LONG direction)
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    if(!fh)
	return DOS_FALSE;
    return (LONG)DoPkt4(fh->fh_Port, ACTION_SEEK, (ULONG)fh,
			(ULONG)offset, (ULONG)direction, 0);
}


static LONG  DeleteFile(const char *filename)
{
    struct DevProc *dev = GetDeviceProc(filename, NULL);
    LONG error = DOS_FALSE;

    if(dev)
    {
	error = DoPkt4(dev->dvp_Port, ACTION_DELETE, dev->dvp_Lock,
		       (ULONG)filename, 0, 0);
	FreeDeviceProc(dev);
    }
    return error;
}


static LONG  Rename(const char *oldName, const char *newName)
{
    LONG error;
    struct DevProc *dev1, *dev2;

    if(!(dev1 = GetDeviceProc(oldName, NULL)))
    {
	SetIoErr(ERROR_OBJECT_NOT_FOUND);
	return DOS_FALSE;
    }
    if(!(dev2 = GetDeviceProc(newName, NULL)))
    {
	FreeDeviceProc(dev1);
	SetIoErr(ERROR_OBJECT_NOT_FOUND);
	return DOS_FALSE;
    }
    if(dev1->dvp_Port == dev2->dvp_Port)
    {
	error = (LONG)DoPkt4(dev1->dvp_Port, ACTION_RENAME,
			     (ULONG)dev1->dvp_Lock, (ULONG)oldName,
			     (ULONG)dev2->dvp_Lock, (ULONG)newName);
    }
    else
    {
	SetIoErr(ERROR_RENAME_ACROSS_DEVICES);
	error = DOS_FALSE;
    }
    FreeDeviceProc(dev1);
    FreeDeviceProc(dev2);
    return error;
}


static APTR  Lock(const char *name, LONG mode)
{
    struct DevProc *dev;
    APTR lukko = NULL;

    if((dev = GetDeviceProc(name, NULL)))
    {
	lukko = (APTR)DoPkt4(dev->dvp_Port, ACTION_LOCK,
			     (ULONG)dev->dvp_Lock, (ULONG)name,
			     mode, 0);
	FreeDeviceProc(dev);
    }
    return lukko;
}


static LONG  UnLock(APTR lock)
{
    struct FileLock *fl = (struct FileLock *)lock;

    if(!lock)
    {
	return DOS_FALSE;
    }
    return DoPkt4(fl->fl_Task, ACTION_UNLOCK, (ULONG)lock, 0, 0, 0);
}


static APTR  DupLock(APTR oldLock)
{
    struct FileLock *fl = (struct FileLock *)oldLock;

    if(!fl)
    {
	return DOS_FALSE;
    }
    return (APTR)DoPkt4(fl->fl_Task, ACTION_DUPLOCK, (ULONG)fl, 0, 0, 0);
}


static LONG  SameLock(APTR lockA, APTR lockB)
{
    struct FileLock *flA = (struct FileLock *)lockA;
    struct FileLock *flB = (struct FileLock *)lockB;

    if(!flA || !flB || flA->fl_Task!=flB->fl_Task)
	return DOS_FALSE;
    return DoPkt4(flA->fl_Task, ACTION_SAME_LOCK, (ULONG)flA, (ULONG)flB, 0,0);
}


static LONG  Examine(APTR lock, struct FileInfoBlock *fib)
{
    struct FileLock *fl = (struct FileLock *)lock;

    if(!fl)
	return DOS_FALSE;
    return DoPkt4(fl->fl_Task, ACTION_EXAMINE, (ULONG)fl, (ULONG)fib, 0, 0);
}


static LONG  ExNext(APTR lock, struct FileInfoBlock *fib)
{
    struct FileLock *fl = (struct FileLock *)lock;

    if(!fl)
	return DOS_FALSE;
    return DoPkt4(fl->fl_Task, ACTION_EXNEXT, (ULONG)fl, (ULONG)fib, 0, 0);
}

static LONG  ExamineEnd(APTR lock, struct FileInfoBlock *fib)
{
    struct FileLock *fl = (struct FileLock *)lock;

    if(!fl)
	return DOS_FALSE;
    return DoPkt4(fl->fl_Task, ACTION_EXAMINE_END, (ULONG)fl, (ULONG)fib,0,0);
}

static APTR  CreateDir(const char *name)
{
    struct DevProc *dev = GetDeviceProc(name, NULL);
    APTR lock = (APTR)DOS_FALSE;

    if(dev)
    {
	lock = (APTR)DoPkt4(dev->dvp_Port, ACTION_CREATE_DIR, dev->dvp_Lock,
			    (ULONG)name, 0, 0);
	FreeDeviceProc(dev);
    }
    return lock;
}


static APTR  CurrentDir(APTR newCurrentDir)
{
    struct FileLock *fl = (struct FileLock *)newCurrentDir, *oldfl;
    struct Process *proc = (struct Process *)FindTask(NULL);

    if(!fl)
	return DOS_FALSE;
    if(DoPkt4(fl->fl_Task, ACTION_CURRENTDIR, (ULONG)fl, 0,0,0))
    {
	/* Valid lock for current directory */
	oldfl = (struct FileLock *)proc->pr_CurrentDir;
	proc->pr_CurrentDir = newCurrentDir;
	proc->pr_FileSystemTask = fl->fl_Task;
	return (APTR)oldfl;
    }
    /* Invalid lock for current directory */
    SetIoErr(ERROR_OBJECT_WRONG_TYPE);
    return NULL;
}


static APTR  ParentDir(APTR lock)
{
    struct FileLock *fl = (struct FileLock *)lock;

    if(!fl)
	return DOS_FALSE;
    return (APTR)DoPkt4(fl->fl_Task, ACTION_PARENT, (ULONG)fl, 0, 0, 0);
}


static APTR  SetComment(const char *filename, const char *comment)
{
    struct DevProc *dev;
    ULONG error = DOS_FALSE;

    if((dev = GetDeviceProc(filename, NULL)))
    {
	error = DoPkt4(dev->dvp_Port, ACTION_SETCOMMENT,
		       (ULONG)dev->dvp_Lock, (ULONG)filename,
		       (ULONG)comment, 0);
	FreeDeviceProc(dev);
    }
    return error;
}


static APTR  SetProtection(const char *filename, ULONG protection)
{
    struct DevProc *dev;
    ULONG error = NULL;

    if((dev = GetDeviceProc(filename, NULL)))
    {
	error = DoPkt4(dev->dvp_Port, ACTION_SETPROTECTION,
		       (ULONG)dev->dvp_Lock, (ULONG)filename,
		       protection, 0);
	FreeDeviceProc(dev);
    }
    return error;
}

static ULONG SetOwner(const char *filename, ULONG ownerinfo)
{
    struct DevProc *dev;
    ULONG error = DOS_FALSE;

    if((dev = GetDeviceProc(filename, NULL)))
    {
	error = (ULONG)DoPkt4(dev->dvp_Port, ACTION_SETOWNER,
			      (ULONG)dev->dvp_Lock, (ULONG)filename,
			      ownerinfo, 0);
	FreeDeviceProc(dev);
    }
    return error;
}

static LONG Format(const char *name, ULONG flags)
{
    char handler[HANDLER_SIZE];

    if(strlen(name)>HANDLER_SIZE)
    {
	SetIoErr(ERROR_INVALID_COMPONENT_NAME);
	return DOS_FALSE;
    }
    strtoupper(handler, name);
    return DoPkt4Safe(handler, ACTION_FORMAT, flags,0,0,0);
}

static LONG Relabel( const char *name, const char *newname )
{
    char handler[HANDLER_SIZE];

    if(strlen(name)>HANDLER_SIZE)
    {
	SetIoErr(ERROR_INVALID_COMPONENT_NAME);
	return DOS_FALSE;
    }
    strtoupper(handler, name);
    return DoPkt4Safe(handler, ACTION_RELABEL, (ULONG)newname,0,0,0);
}


static LONG DiskInfo(APTR lock, struct InfoData *info)
{
    struct FileLock *fl = (struct FileLock *)lock;

    if(!lock)
	return DOS_FALSE;
    return DoPkt4(fl->fl_Task, ACTION_DISKINFO, (ULONG)info, 0, 0, 0);
}

static void  SetIoErr(LONG newval)
{
    ((struct Process *)FindTask(NULL))->pr_Result2 = newval;
}

static LONG  IoErr(void)
{
    return ((struct Process *)FindTask(NULL))->pr_Result2;
}

struct FaultStr
{
    LONG fault_Code;
    char *fault_String;
};
const struct FaultStr Faults[]=
{
    ERROR_NO_FREE_STORE, ": Out of memory.\n",
    ERROR_NO_DEFAULT_DIR, ": No current directory.\n",
    ERROR_OBJECT_IN_USE, ": Object in use.\n",
    ERROR_OBJECT_EXISTS, ": Object exists.\n",
    ERROR_DIR_NOT_FOUND, ": Directory not found.\n",
    ERROR_OBJECT_NOT_FOUND, ": Object not found.\n",
    ERROR_BAD_STREAM_NAME, ": Bad stream name.\n",
    ERROR_OBJECT_TOO_LARGE, ": Object too large.\n",
    ERROR_ACTION_NOT_KNOWN, ": Action not known.\n",
    ERROR_INVALID_COMPONENT_NAME, ": Invalid component name.\n",
    ERROR_INVALID_LOCK, ": Invalid lock.\n",
    ERROR_OBJECT_WRONG_TYPE, ": Object wrong type.\n",
    ERROR_DISK_NOT_VALIDATED, ": Disk not validated.\n",
    ERROR_DISK_WRITE_PROTECTED, ": Disk write protected.\n",
    ERROR_RENAME_ACROSS_DEVICES, ": Rename across devices.\n",
    ERROR_DIRECTORY_NOT_EMPTY, ": Directory not empty.\n",
    ERROR_TOO_MANY_LEVELS, ": Too many levels.\n",
    ERROR_DEVICE_NOT_MOUNTED, ": Device not mounted.\n",
    ERROR_SEEK_ERROR, ": Seek error.\n",
    ERROR_COMMENT_TOO_BIG, ": Comment too big.\n",
    ERROR_DISK_FULL, ": Disk full.\n",
    ERROR_DELETE_PROTECTED, ": Delete protected.\n",
    ERROR_WRITE_PROTECTED, ": Write protected.\n",
    ERROR_READ_PROTECTED, ": Read protected.\n",
    ERROR_EXECUTE_PROTECTED, ": Execute protected.\n",
    ERROR_NO_DISK, ": No disk.\n",
    ERROR_NO_MORE_ENTRIES, ": No more entries.\n",
    ERROR_IS_SOFT_LINK, ": Is a soft link.\n",
    ERROR_OBJECT_LINKED, ": Object is linked.\n",
    ERROR_BAD_HUNK, ": Bad hunk.\n",
    ERROR_NOT_IMPLEMENTED, ": Not implemented.\n",
    ERROR_NOT_A_DOS_DISK, ": Not a dos disk.\n",
    ERROR_PARENT_PROTECTED, ": Parent directory write protected.\n",
    ERROR_NOT_OWNER, ": Not owner.\n",
    -1, NULL
};

static LONG  PrintFault( LONG code, const char *header )
{
    int i=0;
    char temp[80];

    while(Faults[i].fault_Code > 0)
    {
	if(Faults[i].fault_Code == code)
	{
	    Write(Error(), header, strlen(header));
	    Write(Error(), Faults[i].fault_String,
		  strlen(Faults[i].fault_String));
	    return DOS_TRUE;
	}
	i++;
    }
    sprintf(temp, "%s: Unknown error. (%d)\n", header, code);
    Write(Error(), temp, strlen(temp));
    return DOS_FALSE;
}

static long WaitForChar(APTR handle, ULONG time)
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    if(!fh)
	return DOS_FALSE;
    return DoPkt4(fh->fh_Port, ACTION_WAITFORCHAR, (ULONG)fh,
		  time, 0, 0);
}

static LONG IsInteractive( APTR handle )
{
    struct FileHandle *fh = (struct FileHandle *)handle;

    if(!fh)
	return DOS_FALSE;
    if(fh->fh_Pos < 0)
	return DOS_TRUE;
    return DOS_FALSE;
}

static LONG SetFileDate(const char *name, const struct DateStamp *date)
{
    struct DevProc *dev;
    ULONG error = DOS_FALSE;

    if((dev = GetDeviceProc(name, NULL)))
    {
	error = (ULONG)DoPkt4(dev->dvp_Port, ACTION_SET_DATE,
			      (ULONG)dev->dvp_Lock, (ULONG)name,
			      (ULONG)date, 0);
	FreeDeviceProc(dev);
    }
    return error;
}

static LONG IsFileSystem( const char *name )
{
    struct DevProc *dev;
    ULONG error = DOS_FALSE;

    if((dev = GetDeviceProc(name, NULL)))
    {
	error = (ULONG)DoPkt4(dev->dvp_Port, ACTION_IS_FILESYSTEM,
			      0, 0, 0, 0);
	FreeDeviceProc(dev);
    }
    return error;
}


static void Delay(ULONG ticks)
{
    struct TimerReq msg;

    if(ticks<1)
	return;
    msg.tr_Node.ln_Type=NT_DELAY;
    msg.tr_Task= FindTask(NULL);
    msg.tr_Ticks = ticks;
    SendTimerReq(&msg);
    Wait(SIGF_DELAY);	/* wait breaks Forbid/Disable */
    while(msg.tr_Node.ln_Type != NT_REPLYMSG)
	Wait(SIGF_DELAY);
}


static LONG DeleteVar(const char *name, ULONG flags)
{
    struct LocalVar *var = (struct LocalVar *)
	    FindName(&((struct Process *)FindTask(NULL))->pr_LocalVars,
		     (UBYTE *)name);
    while(var)
    {
	if(var->lv_Flags == flags)
	{
	    Remove((struct Node *)var);
	    FreeMem(var->lv_Value, var->lv_Len);
	    FreeMem(var, sizeof(struct LocalVar)+strlen(var->lv_Node.ln_Name)+1);
	    return -1;
	}
	var = (struct LocalVar *)FindName((struct List *)var, (UBYTE *)name);
    }
    return 0;
}

static struct LocalVar *FindVar(const char *name, ULONG flags)
{
    struct LocalVar *var = (struct LocalVar *)
	    FindName(&((struct Process *)FindTask(NULL))->pr_LocalVars,
		     (UBYTE *)name);
    while(var)
    {
	if(var->lv_Flags == flags)
	{
	    return var;
	}
	var = (struct LocalVar *)FindName((struct List *)var, (UBYTE *)name);
    }
    return 0;
}

static LONG GetVar(const char *name, char *buffer, LONG bufsize, ULONG flags)
{
    struct LocalVar *var = (struct LocalVar *)
	    FindName(&((struct Process *)FindTask(NULL))->pr_LocalVars,
		     (UBYTE *)name);
    while(var)
    {
	if(var->lv_Flags == flags)
	{
	    if(var->lv_Len > bufsize)
	    {
		memcpy(buffer, var->lv_Value, bufsize);
		return bufsize+1;
	    }
	    memcpy(buffer, var->lv_Value, var->lv_Len);
	    return var->lv_Len;
	}
	var = (struct LocalVar *)FindName((struct List *)var, (UBYTE *)name);
    }
    return -1;
}

static LONG SetVar(const char *name, const char *buffer, LONG size, ULONG flags)
{
    struct LocalVar *var = FindVar(name, flags);
    char *newbuf;

    if(var)
    {
	if(size != var->lv_Len)
	{
	    newbuf = (char *)AllocMem(size, 0);
	    if(newbuf)
	    {
		FreeMem(var->lv_Value, var->lv_Len);
		var->lv_Value = (UBYTE *)newbuf;
		var->lv_Len = size;
		memcpy(var->lv_Value, buffer, size);
		return size;
	    }
	    return 0;
	}
	else
	{
	    memcpy(var->lv_Value, buffer, size);
	    return size;
	}
    }
    else
    {
	var = (struct LocalVar *)AllocMem(sizeof(struct LocalVar)+
					  strlen(name)+1, MEMF_CLEAR);
	if(var)
	{
	    newbuf = (char *)AllocMem(size, 0);
	    if(newbuf)
	    {
		var->lv_Node.ln_Name = (char *)var+sizeof(struct LocalVar);
		strcpy(var->lv_Node.ln_Name, name);
		var->lv_Flags = flags;
		var->lv_Value = (UBYTE *)newbuf;
		var->lv_Len = size;
		memcpy(var->lv_Value, buffer, size);
		AddTail(&((struct Process *)FindTask(NULL))->pr_LocalVars,
			(struct Node *)var);
		return size;
	    }
	    FreeMem(var, sizeof(struct LocalVar)+strlen(name)+1);
	    return 0;
	}
	else
	{
	    return 0;
	}
    }
}
