#include <exec/exec_protos.h>
#include <dos/dos.h>
#include <dos/dos_protos.h>

#include <stdlib.h>
#include <string.h>

LONG handler(void);

const UWORD kludgejump = 0x4ef9;
const ULONG kludgeaddr = (ULONG)handler;

static int launched = 0;
struct DOSLibrary *DOSBase = NULL;

#define BLOCK_SIZE 256 /* (min) size of one buffer */

#define ENTRY_DIR   1
#define ENTRY_FILE -1

struct DirEntry
{
    struct Node de_Node;
    struct DirEntry *de_Parent;
    struct FileInfoBlock de_FIB;
    struct List de_List;
};

struct Buffer
{
    struct MinNode buf_Node;
    ULONG buf_Pos;  /* Position in the file */
    ULONG buf_Size; /* Size of the buffer */
    ULONG buf_Used; /* Used size in the buffer */
    UBYTE buf_Buffer[2]; /* the buffer itself */
};

struct DirEntry *RootNode = NULL;
struct MsgPort *myport = NULL;
struct DosList *dol = NULL;
ULONG bytesused = 0;

struct DirEntry *InitEntry(struct DosPacket *packet, char *name,
			   LONG type, ULONG prot, ULONG owner);
struct DirEntry *LocateEntry(struct DirEntry *base,char *name,int parent);
struct FileLock *FindLock(ULONG key);
struct FileLock *LockEntry(LONG *err, ULONG key, LONG mode);
void UnLockEntry(struct FileLock *lock);
LONG DeleteEntry(struct DirEntry *dirent);
struct DirEntry *GetBaseDir(ULONG lock);
LONG GetAccess(struct DosPacket *packet, struct FileInfoBlock *fib,
	       ULONG access);

LONG ActionDiskInfo(struct DosPacket *packet);
LONG ActionDelete(struct DosPacket *packet);
LONG ActionCreateDir(struct DosPacket *packet);
LONG ActionLock(struct DosPacket *packet);
LONG ActionUnLock(struct DosPacket *packet);
LONG ActionOpen(struct DosPacket *packet);
LONG ActionClose(struct DosPacket *packet);
LONG ActionRead(struct DosPacket *packet);
LONG ActionWrite(struct DosPacket *packet);
LONG ActionSeek(struct DosPacket *packet);
LONG ActionExamine(struct DosPacket *packet);
LONG ActionExNext(struct DosPacket *packet);
LONG ActionExamineEnd(struct DosPacket *packet);
LONG ActionCurrentDir(struct DosPacket *packet);
LONG ActionSameLock(struct DosPacket *packet);
LONG ActionParentDir(struct DosPacket *packet);
LONG ActionSetComment(struct DosPacket *packet);
LONG ActionSetDate(struct DosPacket *packet);
LONG ActionSetOwner(struct DosPacket *packet);
LONG ActionSetProtection(struct DosPacket *packet);
LONG ActionRename(struct DosPacket *packet);

char *FilePart(char *name)
{
    char *a = name;

    while(*name)
    {
	if(*name==':' || *name=='/')
	    a = name+1;
	name++;
    }
    return a;
}

struct DirEntry *GetBaseDir(ULONG lock)
{
    if(!lock)
	return RootNode;
    return (struct DirEntry *)(((struct FileLock *)lock)->fl_Key);
}

LONG GetAccess(struct DosPacket *packet, struct FileInfoBlock *fib,
	       ULONG access)
{
    struct Task *task = packet->dp_Message.mn_ReplyPort->mp_SigTask;
    UWORD UID = task->tc_UserID, GID = task->tc_GroupID;

    if(fib->fib_OwnerID == UID)
    {
	if((fib->fib_Protection & access)==access)
	    return DOS_TRUE;
    }
    if(fib->fib_GroupID == GID)
    {
	if(((fib->fib_Protection>>4) & access)==access)
	    return DOS_TRUE;
    }
    if(((fib->fib_Protection>>8) & access)==access)
	return DOS_TRUE;
    return DOS_FALSE;
}

LONG IsOwner(struct DosPacket *packet, struct FileInfoBlock *fib)
{
    UWORD UID = ((struct Task *)
		 packet->dp_Message.mn_ReplyPort->mp_SigTask)->tc_UserID;

    if(!UID || UID==fib->fib_OwnerID)
	return DOS_TRUE;
    return DOS_FALSE;
}

struct FileLock *FindLock(ULONG key)
{
    struct FileLock *lock = 
	(struct FileLock *)dol->dol_misc.dol_volume.dol_LockList.lh_Head;

    while(lock->fl_Node.lmn_Next)
    {
	if(lock->fl_Key == key)
	    return lock;
	lock = (struct FileLock *)lock->fl_Node.lmn_Next;
    }
    return NULL;
}

struct FileLock *LockEntry(LONG *err, ULONG key, LONG mode)
{
    struct FileLock *lock = FindLock(key), *newlock;

    if((mode == EXCLUSIVE_LOCK) && lock)
    {
	*err = ERROR_OBJECT_IN_USE;
	return NULL;
    }

    if((mode == SHARED_LOCK) && lock && lock->fl_Access!=SHARED_LOCK)
    {
	*err = ERROR_OBJECT_IN_USE;
	return NULL;
    }

    newlock = (struct FileLock *)AllocMem(sizeof(struct FileLock), MEMF_CLEAR);
    if(!newlock)
    {
	*err = ERROR_NO_FREE_STORE;
	return NULL;
    }

    newlock->fl_Key = key;
    newlock->fl_Access = mode;
    newlock->fl_Task = myport;
    AddTail(&dol->dol_misc.dol_volume.dol_LockList, (struct Node *)newlock);
    return newlock;
}

void UnLockEntry(struct FileLock *lock)
{
    Remove((struct Node *)lock);
    FreeMem(lock, sizeof(struct FileLock));
}

LONG handler()
{
    struct Process *myself = (struct Process *)FindTask(NULL);
    struct DosPacket *packet;
    int done = 0, num;
    char handlername[20];

    /* Detaching */
    if(!launched)
    {
	if((DOSBase = (struct DOSLibrary *)OpenLibrary(DOSL_NAME)))
	{
	    struct SegList *segl = (struct SegList *)myself->pr_SegList;

	    launched = 1;
	    myself->pr_SegList = NULL;
	    if(CreateProc("ram-handler", 40, segl, 2048, "hui" ))
		return RETURN_OK;
	    CloseLibrary((struct Library *)DOSBase);
	    return RETURN_FAIL;
	}
	return RETURN_FAIL;
    }
    /* Ok, we are now running freely.. */

    if(!(myport = CreatePort(NULL)))
	return RETURN_FAIL;

    /* First create a devicenode */
    if((dol = (struct DosList *)AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
    {
	bytesused = sizeof(struct DosList);
	dol->dol_Task = myport;
	dol->dol_Node.ln_Name = handlername;
	dol->dol_Node.ln_Type = DLT_VOLUME;

	NewList(&dol->dol_misc.dol_volume.dol_LockList);
	dol->dol_misc.dol_volume.dol_DiskType = ID_DOS_DISK;

	RootNode = InitEntry(NULL, "RAM", ENTRY_DIR,
			     0x111 * (FIBF_READ | FIBF_WRITE | FIBF_EXECUTE),
			     0);

	/* add to dos list ! */
	Forbid();
	strcpy(handlername, "RAM:");
	num = 0;
#if 0
	while(FindName(&DOSBase->dl_Handlers, (UBYTE *)handlername))
	{
	    strcpy(handlername, "RAM0:");
	    handlername[3] = num+'0';
	    num++;
	    strcpy(RootNode->de_FIB.fib_FileName, "RAM ");
	    RootNode->de_FIB.fib_FileName[3] = num+'0';

	    if(num>9)
	    {
		Permit();
		DeleteEntry(RootNode);
		FreeMem((char *)dol, sizeof(struct DosList));
		return 20;
	    }
	}
#else
	if(FindName(&DOSBase->dl_Handlers, (UBYTE *)handlername))
	{
	    Permit();
	    DeleteEntry(RootNode);
	    FreeMem((char *)dol, sizeof(struct DosList));
	    return 20;
	}
#endif
	AddTail(&DOSBase->dl_Handlers, (struct Node *)dol);
	Permit();

	while(!done)
	{
	    Wait((1L<<myport->mp_SigBit));
	    while((packet = (struct DosPacket *)GetMsg(myport)))
	    {
		switch(packet->dp_Type)
		{
		case ACTION_IS_FILESYSTEM:
		    packet->dp_Result = DOS_TRUE;
		    break;
		case ACTION_EXAMINE:
		    packet->dp_Result = ActionExamine(packet);
		    break;
		case ACTION_EXNEXT:
		    packet->dp_Result = ActionExNext(packet);
		    break;
		case ACTION_EXAMINE_END:
		    packet->dp_Result = ActionExamineEnd(packet);
		    break;
		case ACTION_DELETE:
		    packet->dp_Result = ActionDelete(packet);
		    break;
		case ACTION_CREATE_DIR:
		    packet->dp_Result = ActionCreateDir(packet);
		    break;
		case ACTION_OPEN:
		    packet->dp_Result = ActionOpen(packet);
		    break;
		case ACTION_CLOSE:
		    packet->dp_Result = ActionClose(packet);
		    break;
		case ACTION_READ:
		    packet->dp_Result = ActionRead(packet);
		    break;
		case ACTION_WRITE:
		    packet->dp_Result = ActionWrite(packet);
		    break;
		case ACTION_SEEK:
		    packet->dp_Result = ActionSeek(packet);
		    break;
		case ACTION_LOCK:
		    packet->dp_Result = ActionLock(packet);
		    break;
		case ACTION_UNLOCK:
		    packet->dp_Result = ActionUnLock(packet);
		    break;
		case ACTION_SAME_LOCK:
		    packet->dp_Result = ActionSameLock(packet);
		    break;
		case ACTION_CURRENTDIR:
		    packet->dp_Result = ActionCurrentDir(packet);
		    break;
		case ACTION_PARENT:
		    packet->dp_Result = ActionParentDir(packet);
		    break;
		case ACTION_SETCOMMENT:
		    packet->dp_Result = ActionSetComment(packet);
		    break;
		case ACTION_SETOWNER:
		    packet->dp_Result = ActionSetOwner(packet);
		    break;
		case ACTION_SETPROTECTION:
		    packet->dp_Result = ActionSetProtection(packet);
		    break;
		case ACTION_SET_DATE:
		    packet->dp_Result = ActionSetDate(packet);
		    break;
		case ACTION_RENAME:
		    packet->dp_Result = ActionRename(packet);
		    break;
		case ACTION_DISKINFO:
		    packet->dp_Result = ActionDiskInfo(packet);
		    break;
		case ACTION_DIE:
		    packet->dp_Result = DOS_FALSE;
		    break;
		case ACTION_NIL:
		    packet->dp_Result = DOS_TRUE;
		    break;
		default:
		    packet->dp_Result2 = ERROR_ACTION_NOT_KNOWN;
		    packet->dp_Result = DOS_FALSE;
		}
		ReplyMsg((struct Message *)packet);
	    }
	}
	DeleteEntry(RootNode);
	Remove((struct Node *)dol);
	FreeMem((char *)dol, sizeof(struct DosList));
    }
    DeletePort(myport);
    CloseLibrary((struct Library *)DOSBase);
    return 0;
}


struct DirEntry *LocateEntry(struct DirEntry *base, char *name, int parent)
{
    char namecomp[108];
    int i = 0;
    struct DirEntry *found = NULL;

    while(1)
    {
	if(!*name)
	{
	    /* End of the name */
	    namecomp[i] = '\0';
	    /* Looking for a parent, "" or "." */
	    if(parent || !i || !strcmp(namecomp, "."))
		return base;

	    if(!strcmp(namecomp, ".."))
		return base->de_Parent;

	    if(!base || base->de_FIB.fib_DirEntryType<0)
		return NULL;

	    return (struct DirEntry *)FindName(&base->de_List,
					       (UBYTE *)namecomp);
	}

	if(*name==':')
	{
	    i = 0;
	    /*
	       Handler name can be for this handler, which means we
	       got a lock to the root directory (or NULL). Otherwise
	       the handler name can be an assign and we got a lock
	       to a directory.
	       */
	}
	else if(*name=='/')
	{
	    namecomp[i] = '\0';
	    if(!strcmp(namecomp, "."))
	    {
		/* ./ goes nowhere */
		/* base = base; */
	    }
	    else if(!i || !strcmp(namecomp, ".."))
	    {
		/* "../" or "/" makes us go to the parent directory */
		if(!base->de_Parent)
		    return NULL;
		base = base->de_Parent;
	    }
	    else
	    {
		found = (struct DirEntry *)
		    FindName(&base->de_List, (UBYTE *)namecomp);
		/* entry not found or not a directory */
		if(!found || (found->de_FIB.fib_DirEntryType < 0))
		    return NULL;
		base = found;
	    }
	    i = 0;
	}
	else
	{
	    if(i<106)
		namecomp[i++] = *name;
	}
	name++;
    }
}


struct DirEntry *InitEntry(struct DosPacket *packet, char *name,
			   LONG type, ULONG prot, ULONG owner)
{
    struct DirEntry *dir =
	(struct DirEntry *)AllocMem(sizeof(struct DirEntry), MEMF_CLEAR);

    if(!dir)
	return NULL;

    if(strlen(name)>106)
    {
	if(packet)
	    packet->dp_Result2 = ERROR_INVALID_COMPONENT_NAME;
	FreeMem(dir, sizeof(struct DirEntry));
	return NULL;
    }

    bytesused += sizeof(struct DirEntry);
    NewList(&dir->de_List);
    dir->de_FIB.fib_DiskKey = (LONG)dir;
    dir->de_FIB.fib_DirEntryType = type;
    dir->de_Node.ln_Name = &dir->de_FIB.fib_FileName[0];
    strncpy(dir->de_FIB.fib_FileName, name, 107);
    dir->de_FIB.fib_FileName[107] = '\0';
    dir->de_FIB.fib_Protection = prot;
    dir->de_FIB.fib_Size = 0;
    dir->de_FIB.fib_OwnerID = (owner>>16);
    dir->de_FIB.fib_GroupID = (owner&0xffff);
    DateStamp(&dir->de_FIB.fib_Date);
    return dir;
}

LONG DeleteEntry(struct DirEntry *dirent)
{
    struct FileLock *lock = FindLock((ULONG)dirent);

    if(!dirent || (dirent==RootNode) || lock)
    {
	return ERROR_OBJECT_IN_USE;
    }

    if(dirent->de_FIB.fib_DirEntryType > 0) /* directory */
    {
	if(dirent->de_List.lh_Head->ln_Next)
	    return ERROR_DIRECTORY_NOT_EMPTY;
    }
    else
    {
	struct Buffer *buf;

	while((buf = (struct Buffer *)RemTail(&dirent->de_List)))
	{
	    bytesused -= sizeof(struct Buffer)+buf->buf_Size;
	    FreeMem(buf, buf->buf_Size + sizeof(struct Buffer));
	}
    }
    Remove((struct Node *)dirent);
    FreeMem(dirent, sizeof(struct DirEntry));
    bytesused -= sizeof(struct DirEntry);
    return 0;
}

LONG ActionDiskInfo(struct DosPacket *packet)
{
    struct InfoData *info = (struct InfoData *)packet->dp_Arg1;

    info->id_NumSoftErrors = 0;     /* number of soft errors on disk */
    info->id_UnitNumber = 0;        /* Which unit disk is (was) mounted on */
    info->id_DiskState = ID_VALIDATED;
    info->id_NumBlocksUsed = info->id_NumBlocks = bytesused/BLOCK_SIZE;
    info->id_BytesPerBlock = BLOCK_SIZE;
    info->id_DiskType = ID_DOS_DISK;
    info->id_VolumeNode = NULL;
    info->id_InUse = -1; /* TODO: Number of locks */

    return DOS_TRUE;
}

LONG ActionLock(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }

    return (LONG)LockEntry(&packet->dp_Result2,
			   (ULONG)dirent, (LONG)packet->dp_Arg3);
}

LONG ActionUnLock(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;

    UnLockEntry(fl);
    return DOS_TRUE;
}

LONG ActionOpen(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);
    struct FileHandle *fh;
    struct DirEntry *parent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 1);
    struct Task *task = packet->dp_Message.mn_ReplyPort->mp_SigTask;
    ULONG owner = (task->tc_UserID<<16) | task->tc_GroupID;
    char *name = FilePart((char *)packet->dp_Arg2);

    if(!parent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!strcmp(name, "..") || !strcmp(name, "."))
    {
	packet->dp_Result2 = ERROR_OBJECT_EXISTS;
	return DOS_FALSE;
    }
    if(dirent && dirent->de_FIB.fib_DirEntryType > 0)
    {
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }

    fh = (struct FileHandle *)AllocMem(sizeof(struct FileHandle), MEMF_CLEAR);
    if(!fh)
    {
	packet->dp_Result2 = ERROR_NO_FREE_STORE;
	return DOS_FALSE;
    }

    switch(packet->dp_Arg3) /* Open mode */
    {
    case MODE_OLDFILE:
	if(!dirent)
	{
	    packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	if(!GetAccess(packet, &dirent->de_FIB, FIBF_READ))
	{
	    packet->dp_Result2 = ERROR_READ_PROTECTED;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	fh->fh_Port = myport;
	fh->fh_Arg1 = (ULONG)dirent;
	fh->fh_Buf  = NULL; /* does not point to a buffer */
	fh->fh_Pos  = 0;
	if(fh->fh_Arg2 =
	   (ULONG)LockEntry(&packet->dp_Result2, (ULONG)dirent, SHARED_LOCK))
	{
	    return (LONG)fh;
	}
	FreeMem(fh, sizeof(struct FileHandle));
	return DOS_FALSE;
	break;
    case MODE_NEWFILE:
	if(dirent)
	{
	    if(!GetAccess(packet, &dirent->de_FIB, FIBF_DELETE))
	    {
		packet->dp_Result2 = ERROR_DELETE_PROTECTED;
		FreeMem(fh, sizeof(struct FileHandle));
		return DOS_FALSE;
	    }
	    if((packet->dp_Result2 = DeleteEntry(dirent)))
	    {
		FreeMem(fh, sizeof(struct FileHandle));
		return DOS_FALSE;
	    }
	}
	if(!GetAccess(packet, &parent->de_FIB, FIBF_WRITE))
	{
	    packet->dp_Result2 = ERROR_PARENT_PROTECTED;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	if((dirent =
	    InitEntry(packet, name, ENTRY_FILE,
		      0x111 * (FIBF_DELETE | FIBF_READ | FIBF_WRITE | FIBF_EXECUTE),
		      owner)))
	{
	    dirent->de_Parent = parent;
	    fh->fh_Port = myport;
	    fh->fh_Arg1 = (ULONG)dirent;
	    fh->fh_Buf  = NULL; /* does not point to a buffer */
	    fh->fh_Pos  = 0;
	    if(fh->fh_Arg2 = (ULONG)
	       LockEntry(&packet->dp_Result2, (ULONG)dirent, EXCLUSIVE_LOCK))
	    {
		AddTail(&parent->de_List, (struct Node *)dirent);
		return (LONG)fh;
	    }
	}
	FreeMem(fh, sizeof(struct FileHandle));
	return DOS_FALSE;
	break;
    case MODE_READWRITE:
	if(dirent)
	{
	    if(!GetAccess(packet, &dirent->de_FIB, FIBF_WRITE))
	    {
		packet->dp_Result2 = ERROR_WRITE_PROTECTED;
		FreeMem(fh, sizeof(struct FileHandle));
		return DOS_FALSE;
	    }
	    fh->fh_Port = myport;
	    fh->fh_Arg1 = (ULONG)dirent;
	    fh->fh_Buf  = NULL; /* does not point to a buffer */
	    fh->fh_Pos  = 0;
	    if(fh->fh_Arg2 = (ULONG)
	       LockEntry(&packet->dp_Result2, (ULONG)dirent, EXCLUSIVE_LOCK))
	    {
		AddTail(&parent->de_List, (struct Node *)dirent);
		return (LONG)fh;
	    }
	}
	else
	{
	    if(!GetAccess(packet, &parent->de_FIB, FIBF_WRITE))
	    {
		packet->dp_Result2 = ERROR_PARENT_PROTECTED;
		FreeMem(fh, sizeof(struct FileHandle));
		return DOS_FALSE;
	    }
	    if((dirent =
		InitEntry(packet, FilePart((char *)packet->dp_Arg2),
			  ENTRY_FILE,
			  0x111 * (FIBF_DELETE | FIBF_READ | FIBF_WRITE | FIBF_EXECUTE),
			  owner)))
	    {
		dirent->de_Parent = parent;
		fh->fh_Port = myport;
		fh->fh_Arg1 = (ULONG)dirent;
		fh->fh_Buf  = NULL; /* does not point to a buffer */
		fh->fh_Pos  = 0;
		if(fh->fh_Arg2 = (ULONG)
		   LockEntry(&packet->dp_Result2, (ULONG)dirent, EXCLUSIVE_LOCK))
		{
		    AddTail(&parent->de_List, (struct Node *)dirent);
		    return (LONG)fh;
		}
	    }
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	break;
    default:
	packet->dp_Result2 = ERROR_NOT_IMPLEMENTED;
	FreeMem(fh, sizeof(struct FileHandle));
	return DOS_FALSE;
    }
    FreeMem(fh, sizeof(struct FileHandle));
    return DOS_FALSE;
}

LONG ActionClose(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;

    if(fh->fh_Arg2)
	UnLockEntry((struct FileLock *)fh->fh_Arg2);
    FreeMem(fh, sizeof(struct FileHandle));
    return DOS_TRUE;
}

LONG ActionRead(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    char *buffer = (char *)packet->dp_Arg2;
    struct DirEntry *dirent = (struct DirEntry *)fh->fh_Arg1;
    struct Buffer *buf = (struct Buffer *)fh->fh_Buf;
    ULONG left,loc;
    ULONG size = packet->dp_Arg3;

    if(!GetAccess(packet, &dirent->de_FIB, FIBF_READ))
    {
	packet->dp_Result2 = ERROR_READ_PROTECTED;
	return DOS_FALSE;
    }
    if(!buf)
    {
	buf = (struct Buffer *)dirent->de_List.lh_Head;
    }

    if(!buf->buf_Node.lmn_Next)
    {
	return DOS_FALSE; /* EOF, no characters read */
    }
    while(size && buf->buf_Node.lmn_Next)
    {
	loc = fh->fh_Pos - buf->buf_Pos; /* position inside this buffer */
	left = buf->buf_Used - loc;      /* characters left in this buffer */

	if(size<left)
	    left = size;

	memcpy(buffer, buf->buf_Buffer + loc, left);
	buffer += left;       /* advance in the destination buffer */
	size -= left;         /* not so much to copy anymore */
	fh->fh_Pos += left;   /* position in the file is changed also */

	fh->fh_Buf = (ULONG)buf;
	buf = (struct Buffer *)buf->buf_Node.lmn_Next; /* next buffer */
    }
    packet->dp_Result2 = 0;
    return (LONG)(packet->dp_Arg3 - size);
}

LONG ActionWrite(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    char *buffer = (char *)packet->dp_Arg2;
    struct DirEntry *dirent = (struct DirEntry *)fh->fh_Arg1, *tmpent;
    struct Buffer *buf = (struct Buffer *)fh->fh_Buf;
    ULONG left,loc;
    ULONG size = packet->dp_Arg3;

    if(!GetAccess(packet, &dirent->de_FIB, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }
    if(!buf)
    {
	buf = (struct Buffer *)dirent->de_List.lh_Head;
    }

    /* This buffer is valid and this buffer is also full */
    if(buf->buf_Node.lmn_Next && (fh->fh_Pos == buf->buf_Pos+buf->buf_Size))
    {
	buf = (struct Buffer *)buf->buf_Node.lmn_Next;
	/* next buffer, if any */
    }
    while(size)
    {
	/* Not a valid buffer*/
	if(!buf->buf_Node.lmn_Next)
	{
	    left = BLOCK_SIZE;
	    if(size > left)
	    {
		left = size + BLOCK_SIZE;
	    }
	    buf=(struct Buffer *)AllocMem(sizeof(struct Buffer)+left,0);
	    if(!buf)
	    {
		packet->dp_Result2 = ERROR_NO_FREE_STORE;
		break;
	    }
	    bytesused += sizeof(struct Buffer)+left;
	    buf->buf_Pos = fh->fh_Pos;
	    buf->buf_Used = 0;
	    buf->buf_Size = left; /*leaves 2 bytes unused, but so what? */
	    AddTail(&dirent->de_List, (struct Node *)buf);
	}

	fh->fh_Buf = (ULONG)buf;
	loc = fh->fh_Pos - buf->buf_Pos; /* position inside this buffer */
	left = buf->buf_Size - loc;      /* space left in this buffer */

	if(size<left)
	    left = size;

	memcpy(buf->buf_Buffer+loc, buffer, left);
	if(loc + left > buf->buf_Used)
	    buf->buf_Used = loc + left;

	buffer += left;       /* advance in the source buffer */
	size -= left;         /* not so much to copy anymore */
	fh->fh_Pos += left;   /* position in the file is changed */

	if(fh->fh_Pos > dirent->de_FIB.fib_Size)
	    dirent->de_FIB.fib_Size = fh->fh_Pos;

	buf = (struct Buffer *)buf->buf_Node.lmn_Next; /* next buffer */
    }
    /* Update dates */
    DateStamp(&dirent->de_FIB.fib_Date);
    tmpent = dirent->de_Parent;
    while(tmpent)
    {
	DateStamp(&tmpent->de_FIB.fib_Date);
	tmpent = tmpent->de_Parent;
    }
    return (LONG)(packet->dp_Arg3 - size);
}

LONG ActionSeek(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    LONG offset = (LONG)packet->dp_Arg2;
    LONG mode   = (LONG)packet->dp_Arg3;
    struct DirEntry *dirent = (struct DirEntry *)fh->fh_Arg1;
    ULONG size  = dirent->de_FIB.fib_Size, loc, oldpos = fh->fh_Pos;

    struct Buffer *buf = (struct Buffer *)fh->fh_Buf;

    if(mode==OFFSET_END)
    {
	loc = size + offset;
    }
    else if(mode==OFFSET_BEGINNING)
    {
	loc = offset;
    }
    else if(mode==OFFSET_CURRENT)
    {
	loc = fh->fh_Pos + offset;
    }
    else
    {
	packet->dp_Result2 = ERROR_SEEK_ERROR;
	return DOS_FALSE;
    }
    if(loc < 0 || loc > size)
    {
	packet->dp_Result2 = ERROR_SEEK_ERROR;
	return DOS_FALSE;
    }

    buf = (struct Buffer *)dirent->de_List.lh_Head;
    fh->fh_Buf = NULL;
    fh->fh_Pos = loc;

    while(buf->buf_Node.lmn_Next)
    {
	if(loc >= buf->buf_Pos && loc < buf->buf_Pos + buf->buf_Used)
	{
	    fh->fh_Buf = (ULONG)buf;
	    break;
	}
	buf = (struct Buffer *)buf->buf_Node.lmn_Next; /* next buffer */
    }
    packet->dp_Result2 = (LONG)oldpos;
    return DOS_TRUE;
}


LONG ActionCreateDir(struct DosPacket *packet)
{
    struct DirEntry *parent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 1);
    struct DirEntry *dirent, *tmpent;
    char *name = FilePart((char *)packet->dp_Arg2);
    struct FileLock *fl;
    struct Task *task = packet->dp_Message.mn_ReplyPort->mp_SigTask;
    ULONG owner = (task->tc_UserID<<16) | task->tc_GroupID;

    if(!*name || !parent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(strlen(name)>106)
    {
	packet->dp_Result2 = ERROR_INVALID_COMPONENT_NAME;
	return DOS_FALSE;
    }
    if(!strcmp(name, "..") || !strcmp(name, "."))
    {
	packet->dp_Result2 = ERROR_OBJECT_EXISTS;
	return DOS_FALSE;
    }

    dirent = LocateEntry(parent, name, 0);
    if(dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_EXISTS;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &parent->de_FIB, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_PARENT_PROTECTED;
	return DOS_FALSE;
    }
    if(dirent =
       InitEntry(packet, name, ENTRY_DIR,
		 0x111 * (FIBF_READ | FIBF_WRITE | FIBF_EXECUTE | FIBF_DELETE),
		 owner))
    {
	dirent->de_Parent = parent;
	if(fl = LockEntry(&packet->dp_Result2, (ULONG)dirent, SHARED_LOCK))
	{
	    AddTail(&parent->de_List, (struct Node *)dirent);

	    /* Update dates */
	    DateStamp(&dirent->de_FIB.fib_Date);
	    tmpent = dirent->de_Parent;
	    while(tmpent)
	    {
		DateStamp(&tmpent->de_FIB.fib_Date);
		tmpent = tmpent->de_Parent;
	    }
	    return (LONG)fl;
	}
    }
    packet->dp_Result2 = ERROR_NO_FREE_STORE;
    return DOS_FALSE;
}

LONG ActionDelete(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);
    struct DirEntry *tmpent = dirent->de_Parent;

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &dirent->de_FIB, FIBF_DELETE))
    {
	packet->dp_Result2 = ERROR_DELETE_PROTECTED;
	return DOS_FALSE;
    }
    if((packet->dp_Result2 = DeleteEntry(dirent)))
	return DOS_FALSE;

    /* Update dates */
    while(tmpent)
    {
	DateStamp(&tmpent->de_FIB.fib_Date);
	tmpent = tmpent->de_Parent;
    }
    return DOS_TRUE;
}

LONG ActionExamine(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct FileInfoBlock *fib = (struct FileInfoBlock *)packet->dp_Arg2;

    memcpy(fib, &((struct DirEntry *)fl->fl_Key)->de_FIB,
	   sizeof(struct FileInfoBlock));
    return DOS_TRUE;
}

LONG ActionExNext(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct FileInfoBlock *fib = (struct FileInfoBlock *)packet->dp_Arg2;
    struct DirEntry *dirent = (struct DirEntry *)fib->fib_DiskKey;

    if(((struct DirEntry *)fl->fl_Key)->de_FIB.fib_DirEntryType < 0)
    {
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }
    if(fib->fib_DiskKey == fl->fl_Key)
    {
	/* Get the first file in the directory */
	dirent = (struct DirEntry *)dirent->de_List.lh_Head;
    }
    else
    {
	dirent = (struct DirEntry *)dirent->de_Node.ln_Next;
    }
    if(dirent->de_Node.ln_Next)
    {
	memcpy(fib, &dirent->de_FIB, sizeof(struct FileInfoBlock));
	return DOS_TRUE;
    }
    packet->dp_Result2 = ERROR_NO_MORE_ENTRIES;
    return DOS_FALSE;
}

LONG ActionExamineEnd(struct DosPacket *packet)
{
#if 0
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct FileInfoBlock *fib = (struct FileInfoBlock *)packet->dp_Arg2;

    /* No buffering done yet, so no cleanup to do either */
#endif
    return DOS_TRUE;
}

LONG ActionCurrentDir(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct DirEntry *dirent = (struct DirEntry *)fl->fl_Key;

    if(dirent->de_FIB.fib_DirEntryType != ENTRY_DIR)
    {
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &dirent->de_FIB, FIBF_EXECUTE))
    {
	packet->dp_Result2 = ERROR_EXECUTE_PROTECTED;
	return DOS_FALSE;
    }
    return (LONG)fl;
}

LONG ActionSameLock(struct DosPacket *packet)
{
    struct FileLock *flA = (struct FileLock *)packet->dp_Arg1;
    struct FileLock *flB = (struct FileLock *)packet->dp_Arg2;

    if(flA->fl_Key == flB->fl_Key)
	return DOS_TRUE;
    return DOS_FALSE;
}

LONG ActionParentDir(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct DirEntry *dirent = (struct DirEntry *)fl->fl_Key;

    if(dirent == RootNode)
    {
	return DOS_FALSE;
    }
    if(dirent->de_Parent &&
       (fl = LockEntry(&packet->dp_Result2,
		       (ULONG)dirent->de_Parent, ACCESS_READ)))
    {
	return (LONG)fl;
    }
    return DOS_FALSE;
}

LONG ActionSetComment(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &dirent->de_FIB, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }
    if(strlen((char *)packet->dp_Arg3)>79)
    {
	packet->dp_Result2 = ERROR_COMMENT_TOO_BIG;
	return DOS_FALSE;
    }
    strcpy(dirent->de_FIB.fib_Comment, (char *)packet->dp_Arg3);
    return DOS_TRUE;
}

LONG ActionSetDate(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &dirent->de_FIB, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }
    memcpy((char *)&dirent->de_FIB.fib_Date,
	   (char *)packet->dp_Arg3,
	   sizeof(struct DateStamp));

    return DOS_TRUE;
}


LONG ActionSetOwner(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!IsOwner(packet, &dirent->de_FIB))
    {
	packet->dp_Result2 = ERROR_NOT_OWNER;
	return DOS_FALSE;
    }
    dirent->de_FIB.fib_OwnerID = (packet->dp_Arg3>>16);
    dirent->de_FIB.fib_GroupID = (packet->dp_Arg3 & 0xffff);
    return DOS_TRUE;
}

LONG ActionSetProtection(struct DosPacket *packet)
{
    struct DirEntry *dirent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!IsOwner(packet, &dirent->de_FIB))
    {
	packet->dp_Result2 = ERROR_NOT_OWNER;
	return DOS_FALSE;
    }
    dirent->de_FIB.fib_Protection = packet->dp_Arg3;
    return DOS_TRUE;
}

LONG ActionRename(struct DosPacket *packet)
{
    struct DirEntry *dirent1 = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);
    struct DirEntry *dirent2 = LocateEntry(GetBaseDir(packet->dp_Arg3),
					  (char *)packet->dp_Arg4, 0);
    struct DirEntry *destdir = LocateEntry(GetBaseDir(packet->dp_Arg3),
					  (char *)packet->dp_Arg4, 1);

    if(!dirent1 || !destdir)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(dirent1 == RootNode || !GetAccess(packet, &dirent1->de_FIB, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }
    if(dirent1 == dirent2)
    {
	/* Nothing to do - the same file/dir */
	return DOS_TRUE;
    }

    /* TODO: check if the parent dir is being examined */
    if(!dirent2)
    {
	if(!GetAccess(packet, &destdir->de_FIB, FIBF_WRITE))
	{
	    packet->dp_Result2 = ERROR_PARENT_PROTECTED;
	    return DOS_FALSE;
	}
	if(dirent1->de_Parent == destdir)
	{
	    /* Rename inside a directory */
	    strncpy(dirent1->de_FIB.fib_FileName,
		    FilePart((char *)packet->dp_Arg4), 107);
	    dirent1->de_FIB.fib_FileName[107] = '\0';
	    return DOS_TRUE;
	}

	Remove((struct Node *)dirent1);
	strncpy(dirent1->de_FIB.fib_FileName,
		FilePart((char *)packet->dp_Arg4), 107);
	dirent1->de_FIB.fib_FileName[107] = '\0';
	AddTail(&destdir->de_List, (struct Node *)dirent1);
	return DOS_TRUE;
    }
    if(dirent2->de_FIB.fib_DirEntryType > 0)
    {
	/* TODO: put the file IN the directory */
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &dirent2->de_FIB, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, &dirent2->de_FIB, FIBF_DELETE))
    {
	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }
    if((packet->dp_Result2 = DeleteEntry(dirent2)))
    {
	return DOS_FALSE;
    }
    Remove((struct Node *)dirent1);
    strncpy(dirent1->de_FIB.fib_FileName,
	    FilePart((char *)packet->dp_Arg4), 107);
    dirent1->de_FIB.fib_FileName[107] = '\0';
    AddTail(&destdir->de_List, (struct Node *)dirent1);
    return DOS_TRUE;
}
