#include <exec/exec_protos.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <dos/dos_protos.h>

#include <stdlib.h>
#include <string.h>

LONG handler(void);
const UWORD kludgejump = 0x4ef9;
const ULONG kludgeaddr = (ULONG)handler;

static int FlashWriteWord(UWORD *address, ULONG value)
{
    ULONG chipmask  = (ULONG)address & 0xff0000;

    if((*(volatile UWORD *)address & value) != value)
	return -1; /* Not possible */

    /* FLASH Byte program command */
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0xaaaa;
    *((volatile UWORD *)(chipmask | 0x5554)) = 0x5555;
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0xa0a0;
    /*
       It IS possible to write individual bytes, but because it
       isn't any faster, we reuse the code and save some bytes.
     */
    *((volatile UWORD *)address) = value;

    /* wait for 14 us (minimum) == 280cycles * 50ns/cycle */
    /* This IS busy-waiting, but the time is too short.. TODO: something? */
    while(*((volatile UWORD *)address) != value);
    return 0;
}

static int FlashWriteLong(ULONG *address, ULONG value)
{
    if(FlashWriteWord((UWORD *)address, (value>>16)))
	return -1;
    if(FlashWriteWord(((UWORD *)address)+1, (value & 0xffff)))
	return -1;
    return 0;
}

static void FlashReadReset(ULONG address)
{
    ULONG chipmask  = address & 0xff0000;

    /* FLASH Read/Reset command */
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0xaaaa;
    *((volatile UWORD *)(chipmask | 0x5554)) = 0x5555;
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0xf0f0;
}

static void FlashEraseChip(ULONG address)
{
    ULONG chipmask  = address & 0xff0000;

    /* Erase the chip pair that has the specified address */
    /* FLASH Chip Erase command */
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0xaaaa;
    *((volatile UWORD *)(chipmask | 0x5554)) = 0x5555;
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0x8080;
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0xaaaa;
    *((volatile UWORD *)(chipmask | 0x5554)) = 0x5555;
    *((volatile UWORD *)(chipmask | 0xaaaa)) = 0x1010;
}

static void FlashErase(void)
{
    /* Flash chip pair addresses: 840000, 900000, a00000, b00000 */
    FlashEraseChip(0x840000);
    FlashEraseChip(0x900000);
    FlashEraseChip(0xa00000);
    FlashEraseChip(0xb00000);

    /* Wait for the erases to complete */
    while(1)
    {
	if(*((volatile UWORD *)0x840000) == 0xffff)
	   break;
	/* It takes >10s to erase the chips, wait 1s and check again */
	Delay(100);
    }
    while(1)
    {
	if(*((volatile UWORD *)0x900000) == 0xffff)
	   break;
	/* It should take the same amount of time to erase all chips */
	Delay(10);
    }
    while(1)
    {
	if(*((volatile UWORD *)0xa00000) == 0xffff)
	   break;
	Delay(10);
    }
    while(1)
    {
	if(*((volatile UWORD *)0xb00000) == 0xffff)
	   break;
	Delay(10);
    }
    FlashReadReset(0x840000);
    FlashReadReset(0x900000);
    FlashReadReset(0xa00000);
    FlashReadReset(0xb00000);
}


struct BootBlock
{
    ULONG bb_DiskType;
    ULONG bb_CheckSum;
    ULONG bb_LowBlock;
    ULONG bb_HiBlock;
    ULONG bb_BlockSize;
    ULONG bb_RootBlock;
    ULONG bb_BitMapBlock; /* BitMapBlocks are contiguous */
    ULONG bb_BitMapBlocks;
};

struct DirEntryBlock
{
    LONG  deb_Type;     /* 2=RootBlock, 1=Dir, -1=File */
    ULONG deb_CheckSum; /* Sum of the whole block becomes 0 */

    ULONG deb_HashChain;    /* Next block in this chain */
    ULONG deb_ParentDir;    /* NULL for RootBlock */
    struct DateStamp deb_ChangeDate;
    char  deb_Comment[80];   /* May be non-NUL-terminated! */
    char  deb_FileName[108]; /* Currently NUL-terminated (max 107 chars) */
    ULONG deb_Owner;        /* UserID and GroupID of the owner */
    ULONG deb_Protect;      /* Protection bits */
    ULONG deb_Extension;    /* Points to extension block (file) */
                            /*           bit map block (rootblock) */
    ULONG deb_Blocks;       /* Blocks used by this file/dir */
    ULONG deb_ByteSize;     /* Bytes used by this file */
    ULONG deb_Reserved;

    /* 60 longs this far */
    ULONG deb_HashTable[4 /* for 256-byte blocksize */];
};

struct DataBlock
{
    UBYTE db_Data[256 /* for 256-byte blocksize */];
};

struct BitMapBlock
{
    ULONG bmb_BitMap[64];/* for 256-byte blocksize */
};

struct FileExtBlock
{
    LONG  feb_Type;      /* -129 */
    ULONG feb_Extension; /* next Extension block or NULL */
    ULONG feb_Data[62 /* for 256-byte blocksize */];
};


static int launched = 0;
struct DOSLibrary *DOSBase;
struct DosList *dol;
struct MsgPort *myport;
struct BootBlock *bb;

#define ENTRY_DIR   1
#define ENTRY_FILE -1

UWORD CalcHashNoCase(char *name);
char *FilePart(char *name);

struct DirEntryBlock *GetBaseDir(ULONG lock);
LONG GetAccess(struct DosPacket *packet, struct DirEntryBlock *deb,
	       ULONG access);
LONG IsOwner(struct DosPacket *packet, struct DirEntryBlock *deb);
LONG IsDirEmpty(struct DirEntryBlock *dir);

#define ALLOC_INFO  0 /* For DirEntryBlocks, FileExtBlocks etc. */
#define ALLOC_DATA  1 /* For file data blocks */
LONG AllocBlock(int type);
void FreeBlock(ULONG block);
ULONG CountBlocks(void);

struct DirEntryBlock *GetBlock(ULONG blocknum);
void WriteBlock(ULONG block, ULONG *source, int size);
LONG WriteFileBlock(struct DirEntryBlock *file, ULONG blockoffset, ULONG *block);
LONG TellFileBlock(struct DirEntryBlock *file, ULONG blockoffset);
struct DirEntryBlock *CopyBlock(struct DirEntryBlock *block);
void FreeCopyBlock(struct DirEntryBlock *block);

struct FileLock *FindLock(ULONG key);
struct FileLock *LockEntry(LONG *error, ULONG key, LONG mode);
void UnLockEntry(struct FileLock *lock);
void InitDirEntry(struct DirEntryBlock *deb, LONG type,
		  struct DirEntryBlock *parent, char *name, ULONG owner);
void InitFileInfoBlock(struct FileInfoBlock *fib,
		       struct DirEntryBlock *deb, ULONG block);
void AddDirEntry(struct DirEntryBlock *parent, char *name, ULONG block);
LONG DeleteDirEntry(struct DirEntryBlock *file, ULONG key);


LONG ActionFormat(struct DosPacket *packet);
LONG ActionDiskInfo(struct DosPacket *packet);
LONG ActionDelete(struct DosPacket *packet);
LONG ActionCreateDir(struct DosPacket *packet);
LONG ActionLock(struct DosPacket *packet);
LONG ActionUnLock(struct DosPacket *packet);
LONG ActionOpen(struct DosPacket *packet);
LONG ActionClose(struct DosPacket *packet);
LONG ActionRead(struct DosPacket *packet);
LONG ActionWrite(struct DosPacket *packet);
LONG ActionSeek(struct DosPacket *packet);
LONG ActionExamine(struct DosPacket *packet);
LONG ActionExNext(struct DosPacket *packet);
LONG ActionExamineEnd(struct DosPacket *packet);
LONG ActionCurrentDir(struct DosPacket *packet);
LONG ActionSameLock(struct DosPacket *packet);
LONG ActionParentDir(struct DosPacket *packet);
LONG ActionSetComment(struct DosPacket *packet);
LONG ActionSetOwner(struct DosPacket *packet);
LONG ActionSetProtection(struct DosPacket *packet);
LONG ActionRename(struct DosPacket *packet);



LONG handler()
{
    struct Process *myself = (struct Process *)FindTask(NULL);
    struct DosPacket *packet;
    int done = 0;
    struct DirEntryBlock *rootblock;
    LONG err;

    /* Detaching */
    if(!launched)
    {
	if((DOSBase = (struct DOSLibrary *)OpenLibrary(DOSL_NAME)))
	{
	    struct SegList *segl = (struct SegList *)myself->pr_SegList;

	    launched = 1;
	    myself->pr_SegList = NULL;
	    if(CreateProc("rom-handler", 40, segl, 2048, "hui" ))
		return RETURN_OK;
	    CloseLibrary((struct Library *)DOSBase);
	    return RETURN_FAIL;
	}
	return RETURN_FAIL;
    }
    /* Ok, we are now running freely.. */

    if(!(myport = CreatePort(NULL)))
	return RETURN_FAIL;

    Forbid();
    if(FindName(&DOSBase->dl_Handlers, (UBYTE *)"ROM:"))
    {
	Permit();
	return RETURN_FAIL;
    }

    /* First create a devicenode */
    bb = (struct BootBlock *)0x840000;
    if((dol = (struct DosList *)AllocMem(sizeof(struct DosList), MEMF_CLEAR)))
    {
	dol->dol_Task = myport;
	dol->dol_Node.ln_Name = "ROM:";
	dol->dol_Node.ln_Type = DLT_VOLUME;

	NewList(&dol->dol_misc.dol_volume.dol_LockList);
	if(bb->bb_DiskType == ID_DOS_DISK)
	{
	    dol->dol_misc.dol_volume.dol_DiskType = ID_DOS_DISK;
	    rootblock = GetBlock(bb->bb_RootBlock);
	    dol->dol_Lock = (APTR)LockEntry(&err, bb->bb_RootBlock, ACCESS_READ);
	}
	else
	{
	    dol->dol_misc.dol_volume.dol_DiskType = ID_UNREADABLE_DISK;
	}

	/* add to dos list ! */
	AddTail(&DOSBase->dl_Handlers, (struct Node *)dol);
	Permit();

	/* To be certain that the Flashs are in read mode */
	FlashReadReset(0x840000);
	FlashReadReset(0x900000);
	FlashReadReset(0xa00000);
	FlashReadReset(0xb00000);

	while(!done)
	{
	    Wait((1L<<myport->mp_SigBit));
	    while((packet = (struct DosPacket *)GetMsg(myport)))
	    {
		if(bb->bb_DiskType != ID_DOS_DISK)
		{
		    if(packet->dp_Type != ACTION_FORMAT)
		    {
			packet->dp_Result2 = ERROR_NOT_A_DOS_DISK;
			packet->dp_Result = DOS_FALSE;
		    }
		    else
		    {
			packet->dp_Result = ActionFormat(packet);
			if(bb->bb_DiskType == ID_DOS_DISK)
			{
			    dol->dol_misc.dol_volume.dol_DiskType =
				ID_DOS_DISK;
			    rootblock = GetBlock(bb->bb_RootBlock);
			    dol->dol_Lock =
				(APTR)LockEntry(&err, bb->bb_RootBlock, ACCESS_READ);
			}
			else
			{
			    dol->dol_misc.dol_volume.dol_DiskType =
				ID_UNREADABLE_DISK;
			}
			NewList(&dol->dol_misc.dol_volume.dol_LockList);
		    }
		}
		else switch(packet->dp_Type)
		{
		case ACTION_IS_FILESYSTEM:
		    packet->dp_Result = DOS_TRUE;
		    break;
		case ACTION_FORMAT:
		    packet->dp_Result = ActionFormat(packet);
		    if(bb->bb_DiskType == ID_DOS_DISK)
		    {
			dol->dol_misc.dol_volume.dol_DiskType =
			    ID_DOS_DISK;
			rootblock = GetBlock(bb->bb_RootBlock);
			dol->dol_Lock =
			    (APTR)LockEntry(&err, bb->bb_RootBlock, ACCESS_READ);
		    }
		    else
		    {
			dol->dol_misc.dol_volume.dol_DiskType =
			    ID_UNREADABLE_DISK;
		    }
		    NewList(&dol->dol_misc.dol_volume.dol_LockList);
		    break;
		case ACTION_EXAMINE:
		    packet->dp_Result = ActionExamine(packet);
		    break;
		case ACTION_EXNEXT:
		    packet->dp_Result = ActionExNext(packet);
		    break;
		case ACTION_EXAMINE_END:
		    packet->dp_Result = ActionExamineEnd(packet);
		    break;
		case ACTION_DELETE:
		    packet->dp_Result = ActionDelete(packet);
		    break;
		case ACTION_CREATE_DIR:
		    packet->dp_Result = ActionCreateDir(packet);
		    break;
		case ACTION_OPEN:
		    packet->dp_Result = ActionOpen(packet);
		    break;
		case ACTION_CLOSE:
		    packet->dp_Result = ActionClose(packet);
		    break;
		case ACTION_READ:
		    packet->dp_Result = ActionRead(packet);
		    break;
		case ACTION_WRITE:
		    packet->dp_Result = ActionWrite(packet);
		    break;
		case ACTION_SEEK:
		    packet->dp_Result = ActionSeek(packet);
		    break;
		case ACTION_LOCK:
		    packet->dp_Result = ActionLock(packet);
		    break;
		case ACTION_UNLOCK:
		    packet->dp_Result = ActionUnLock(packet);
		    break;
		case ACTION_SAME_LOCK:
		    packet->dp_Result = ActionSameLock(packet);
		    break;
		case ACTION_CURRENTDIR:
		    packet->dp_Result = ActionCurrentDir(packet);
		    break;
		case ACTION_PARENT:
		    packet->dp_Result = ActionParentDir(packet);
		    break;
		case ACTION_SETCOMMENT:
		    packet->dp_Result = ActionSetComment(packet);
		    break;
		case ACTION_SETOWNER:
		    packet->dp_Result = ActionSetOwner(packet);
		    break;
		case ACTION_SETPROTECTION:
		    packet->dp_Result = ActionSetProtection(packet);
		    break;
		case ACTION_RENAME:
		    packet->dp_Result = ActionRename(packet);
		    break;
		case ACTION_DISKINFO:
		    packet->dp_Result = ActionDiskInfo(packet);
		    break;
		case ACTION_DIE:
		    packet->dp_Result = DOS_FALSE;
		    break;
		case ACTION_NIL:
		    packet->dp_Result = DOS_TRUE;
		    break;
		default:
		    packet->dp_Result2 = ERROR_ACTION_NOT_KNOWN;
		    packet->dp_Result = DOS_FALSE;
		}
		ReplyMsg((struct Message *)packet);
	    }
	}
	Remove((struct Node *)dol);
	FreeMem((char *)dol, sizeof(struct DosList));
    }
    else
    {
	Permit();
    }
    DeletePort(myport);
    CloseLibrary((struct Library *)DOSBase);
    return 0;
}

LONG ActionDiskInfo(struct DosPacket *packet)
{
    struct InfoData *info = (struct InfoData *)packet->dp_Arg1;

    info->id_NumSoftErrors = 0;     /* number of soft errors on disk */
    info->id_UnitNumber = 0;        /* Which unit disk is (was) mounted on */
    info->id_DiskState = ID_VALIDATED;
    info->id_NumBlocks = bb->bb_HiBlock-bb->bb_LowBlock+1;
    info->id_NumBlocksUsed = CountBlocks();
    info->id_BytesPerBlock = bb->bb_BlockSize;
    info->id_DiskType = bb->bb_DiskType;
    info->id_VolumeNode = (APTR)dol;
    info->id_InUse = -1; /* TODO: Number of locks */

    return DOS_TRUE;
}

UWORD CalcHashNoCase(char *name)
{
    UWORD hash = strlen(name);

    while(*name)
    {
	hash = ((UWORD)(13*hash) + (UWORD)*name) & 0x7ff;
	name++;
    }
    return hash;
}

char *FilePart(char *name)
{
    char *a = name;

    while(*name)
    {
	if(*name==':' || *name=='/')
	    a = name+1;
	name++;
    }
    return a;
}


struct DirEntryBlock *GetBaseDir(ULONG lock)
{
    if(!lock)
	return GetBlock(bb->bb_RootBlock);
    return GetBlock(((struct FileLock *)lock)->fl_Key);
}

void InitDirEntry(struct DirEntryBlock *deb, LONG type,
		  struct DirEntryBlock *parent, char *name,
		  ULONG owner)
{
    memset(deb, 0xff, bb->bb_BlockSize);
    deb->deb_Type = type;
    deb->deb_ParentDir = (ULONG)parent/bb->bb_BlockSize;
    DateStamp(&deb->deb_ChangeDate);
    strcpy(deb->deb_FileName, name);
    deb->deb_Owner = owner;

    /* Leave 'unused' entries to 0xff */
    /* Those entries will be filled later */
    /*deb->deb_Protect = 0xfff;*/
    /*deb->deb_HashChain = 0; */
    /*deb->deb_Comment[0] = '\0';*/
    /*deb->deb_Extension = 0;*/
    /*deb->deb_Blocks = 1;*/
    /*deb->deb_ByteSize = 0; Do NOT set this !! */
}

void AddDirEntry(struct DirEntryBlock *parent, char *name, ULONG block)
{
    ULONG hash, hashsize = (bb->bb_BlockSize/4)-60;

    /* Now add it to the TAIL (for ROM) of the hash chain */
    hash = CalcHashNoCase(name) % hashsize;

    if(!~parent->deb_HashTable[hash])
    {
	FlashWriteLong(&parent->deb_HashTable[hash], block);
    }
    else
    {
	parent = GetBlock(parent->deb_HashTable[hash]);
	while(parent->deb_HashChain)
	{
	    if(!~parent->deb_HashChain)
	    {
		FlashWriteLong(&parent->deb_HashChain, block);
		break;
	    }
	    parent = GetBlock(parent->deb_HashChain);
	}
    }
}

LONG DeleteDirEntry(struct DirEntryBlock *file, ULONG key)
{
    struct FileLock *lock = FindLock(key);

    if(lock)
    {
	return ERROR_OBJECT_IN_USE;
    }
    if(file->deb_Type>0 && !IsDirEmpty(file))
    {
	return ERROR_DIRECTORY_NOT_EMPTY;
    }
    FlashWriteLong((ULONG *)&file->deb_Type, 0);
    return 0;
}

void InitFileInfoBlock(struct FileInfoBlock *fib,
		       struct DirEntryBlock *deb, ULONG block)
{
    char *name = deb->deb_FileName;
    ULONG protect = deb->deb_Protect;

    /* For future enhancement */
    while(!*name)
	name++;

    if(!~protect) /* No specific protections set */
    {
	if(deb->deb_Type<0)
	    protect = 0xfff; /* drwe */
	else
	    protect = 0xfff; /* drwe */
    }

    fib->fib_DiskKey = block;
    fib->fib_DirEntryType = deb->deb_Type;
    fib->fib_EntryType = deb->deb_Type;
    strcpy(fib->fib_FileName, name);
    fib->fib_Protection = protect;
    fib->fib_Size = deb->deb_ByteSize;
    fib->fib_NumBlocks = deb->deb_Blocks;
    fib->fib_Date = deb->deb_ChangeDate;
    if((UBYTE)deb->deb_Comment[0] != 0xff)
    {
	strncpy(fib->fib_Comment, deb->deb_Comment, 80);
	fib->fib_Comment[79] = '\0';
    }
    else
	fib->fib_Comment[0] = '\0';

    fib->fib_OwnerID = (deb->deb_Owner>>16);
    fib->fib_GroupID = deb->deb_Owner&0xffff;
}


LONG GetAccess(struct DosPacket *packet, struct DirEntryBlock *deb,
	       ULONG access)
{
    struct Task *task = packet->dp_Message.mn_ReplyPort->mp_SigTask;
    UWORD UID = task->tc_UserID, GID = task->tc_GroupID;
    ULONG protect = deb->deb_Protect;

    if(!~protect) /* No specific protections set */
    {
	if(deb->deb_Type<0)
	    protect = 0xfff; /* drwe */
	else
	    protect = 0xfff; /* drwe */
    }

    if((deb->deb_Owner>>16) == UID)
    {
	if((protect & access)==access)
	    return DOS_TRUE;
    }
    if(deb->deb_Owner&0xffff == GID)
    {
	if(((protect>>4) & access)==access)
	    return DOS_TRUE;
    }
    if(((protect>>8) & access)==access)
	return DOS_TRUE;
    return DOS_FALSE;
}

LONG IsOwner(struct DosPacket *packet, struct DirEntryBlock *deb)
{
    UWORD UID = ((struct Task *)packet->dp_Message.mn_ReplyPort->mp_SigTask)
		->tc_UserID;

    if(!UID || UID==(deb->deb_Owner>>16))
	return DOS_TRUE;
    return DOS_FALSE;
}

LONG IsDirEmpty(struct DirEntryBlock *dir)
{
    ULONG hashsize = (bb->bb_BlockSize/4)-60, i;
    struct DirEntryBlock *deb;

    for(i=0;i<hashsize;i++)
    {
	deb = GetBlock(dir->deb_HashTable[i]);
	while(deb)
	{
	    if(deb->deb_Type)
		return DOS_FALSE;
	    deb = GetBlock(deb->deb_HashChain);
	}
    }
    return DOS_TRUE;
}


LONG AllocBlock(int down)
{
    ULONG entries = (bb->bb_HiBlock - bb->bb_LowBlock)/32 + 1;
    ULONG poslong, posval, posbit;
    struct BitMapBlock *bmb = (struct BitMapBlock *)GetBlock(bb->bb_BitMapBlock);

    if(down) /* For Data blocks */
    {
	poslong = (bb->bb_RootBlock - bb->bb_LowBlock)/32-1;

	while(poslong && !bmb->bmb_BitMap[poslong])
	    poslong--;
	if(!bmb->bmb_BitMap[poslong])
	{
	    poslong = (bb->bb_HiBlock - bb->bb_LowBlock)/32;
	    while(poslong && !bmb->bmb_BitMap[poslong])
		poslong--;
	    if(!bmb->bmb_BitMap[poslong])
		return -1;
	}
	posval = bmb->bmb_BitMap[poslong];
	posbit = 31;
	while(!(posval&(1U<<31)))
	{
	    posbit--;
	    posval <<= 1;
	}
    }
    else
    {
	poslong = (bb->bb_RootBlock - bb->bb_LowBlock)/32;

	while(poslong<entries && !bmb->bmb_BitMap[poslong])
	    poslong++;
	if(poslong==entries)
	{
	    poslong = (bb->bb_RootBlock - bb->bb_LowBlock)/32 - 1;
	    while(poslong && !bmb->bmb_BitMap[poslong])
		poslong--;
	    if(!bmb->bmb_BitMap[poslong])
		return -1;

	    posval = bmb->bmb_BitMap[poslong];
	    posbit = 31;
	    while(!(posval&(1U<<31)))
	    {
		posbit--;
		posval <<= 1;
	    }
	}
	else
	{
	    posval = bmb->bmb_BitMap[poslong];
	    posbit = 0;
	    while(!(posval&1))
	    {
		posbit++;
		posval >>= 1;
	    }
	}
    }
    FlashWriteLong(&bmb->bmb_BitMap[poslong],
		   bmb->bmb_BitMap[poslong] & ~(1U<<posbit));
    return (LONG)(bb->bb_LowBlock + poslong*32 + posbit);
}

void FreeBlock(ULONG block)
{
    ULONG poslong = (block - bb->bb_LowBlock)/32;
    ULONG posbit  = (block - bb->bb_LowBlock)%32;
    struct BitMapBlock *bmb = (struct BitMapBlock *)GetBlock(bb->bb_BitMapBlock);

    if(block < bb->bb_LowBlock || block > bb->bb_HiBlock)
	return;

    FlashWriteLong(&bmb->bmb_BitMap[poslong],
		   bmb->bmb_BitMap[poslong] | (1U<<posbit));
}

static const char counttable[256]=
{
/* 00000000 00000001 00000010 00000011 00000100 00000101 00000110 00000111 */
	8,	7,	7,	6,	7,	6,	6,	5,
/* 00001000 00001001 00001010 00001011 00001100 00001101 00001110 00001111 */
	7,	6,	6,	5,	6,	5,	5,	4,
/* 00010000 00010001 00010010 00010011 00010100 00010101 00010110 00010111 */
	7,	6,	6,	5,	6,	5,	5,	4,
/* 00011000 00011001 00011010 00011011 00011100 00011101 00011110 00011111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 00100000 00100001 00100010 00100011 00100100 00100101 00100110 00100111 */
	7,	6,	6,	5,	6,	5,	5,	4,
/* 00101000 00101001 00101010 00101011 00101100 00101101 00101110 00101111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 00110000 00110001 00110010 00110011 00110100 00110101 00110110 00110111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 00111000 00111001 00111010 00111011 00111100 00111101 00111110 00111111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 01000000 01000001 01000010 01000011 01000100 01000101 01000110 01000111 */
	7,	6,	6,	5,	6,	5,	5,	4,
/* 01001000 01001001 01001010 01001011 01001100 01001101 01001110 01001111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 01010000 01010001 01010010 01010011 01010100 01010101 01010110 01010111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 01011000 01011001 01011010 01011011 01011100 01011101 01011110 01011111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 01100000 01100001 01100010 01100011 01100100 01100101 01100110 01100111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 01101000 01101001 01101010 01101011 01101100 01101101 01101110 01101111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 01110000 01110001 01110010 01110011 01110100 01110101 01110110 01110111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 01111000 01111001 01111010 01111011 01111100 01111101 01111110 01111111 */
	4,	3,	3,	2,	3,	2,	2,	1,
/* 10000000 10000001 10000010 10000011 10000100 10000101 10000110 10000111 */
	7,	6,	6,	5,	6,	5,	5,	4,
/* 10001000 10001001 10001010 10001011 10001100 10001101 10001110 10001111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 10010000 10010001 10010010 10010011 10010100 10010101 10010110 10010111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 10011000 10011001 10011010 10011011 10011100 10011101 10011110 10011111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 10100000 10100001 10100010 10100011 10100100 10100101 10100110 10100111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 10101000 10101001 10101010 10101011 10101100 10101101 10101110 10101111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 10110000 10110001 10110010 10110011 10110100 10110101 10110110 10110111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 10111000 10111001 10111010 10111011 10111100 10111101 10111110 10111111 */
	4,	3,	3,	2,	3,	2,	2,	1,
/* 11000000 11000001 11000010 11000011 11000100 11000101 11000110 11000111 */
	6,	5,	5,	4,	5,	4,	4,	3,
/* 11001000 11001001 11001010 11001011 11001100 11001101 11001110 11001111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 11010000 11010001 11010010 11010011 11010100 11010101 11010110 11010111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 11011000 11011001 11011010 11011011 11011100 11011101 11011110 11011111 */
	4,	3,	3,	2,	3,	2,	2,	1,
/* 11100000 11100001 11100010 11100011 11100100 11100101 11100110 11100111 */
	5,	4,	4,	3,	4,	3,	3,	2,
/* 11101000 11101001 11101010 11101011 11101100 11101101 11101110 11101111 */
	4,	3,	3,	2,	3,	2,	2,	1,
/* 11110000 11110001 11110010 11110011 11110100 11110101 11110110 11110111 */
	4,	3,	3,	2,	3,	2,	2,	1,
/* 11111000 11111001 11111010 11111011 11111100 11111101 11111110 11111111 */
	3,	2,	2,	1,	2,	1,	1,	0
};

ULONG CountBlocks()
{
    ULONG blocks = bb->bb_HiBlock - bb->bb_LowBlock + 1;
    ULONG poslong, posval, longs = (blocks/32)+1, usedblocks = 0;
    struct BitMapBlock *bmb = (struct BitMapBlock *)GetBlock(bb->bb_BitMapBlock);

    for(poslong=0;poslong<longs;poslong++)
    {
	posval = bmb->bmb_BitMap[poslong];
	if(!posval)
	{
	    usedblocks += 32;
	}
	else if(~posval)
	{
	    usedblocks += counttable[posval&0xff] +
		counttable[(posval>>8)&0xff] +
		counttable[(posval>>16)&0xff] +
		counttable[(posval>>24)&0xff];
	}
    }
    return usedblocks;
}

struct DirEntryBlock *GetBlock(ULONG blocknum)
{
    if(!~blocknum)
	return NULL;
    return (struct DirEntryBlock *)(blocknum * bb->bb_BlockSize);
}

void WriteBlock(ULONG block, ULONG *source, int size)
{
    int i;
    ULONG *dest = (ULONG *)(block * bb->bb_BlockSize);

    for(i=0;i<size;i+=4)
    {
	if(*dest != *source)
	    FlashWriteLong(dest, *source);
	source++;
	dest++;
    }
}

LONG WriteFileBlock(struct DirEntryBlock *file, ULONG blockoffset, ULONG *block)
{
    ULONG block2, debsize = (bb->bb_BlockSize/4)-60,
	  newblk, febsize = (bb->bb_BlockSize/4)-2;
    struct FileExtBlock *feb;

    newblk = AllocBlock(ALLOC_DATA);
    if(newblk == -1)
	return ERROR_DISK_FULL;

    WriteBlock(newblk, (ULONG *)block, (int)bb->bb_BlockSize);
    if(blockoffset < debsize)
    {
	/* There is a pointer in the DirEntryBlock */
	if(!~file->deb_HashTable[blockoffset])
	{
	    FlashWriteLong(&file->deb_HashTable[blockoffset],
			   newblk);
	    return 0;
	}
	else
	{
	    /* This block already written */
	    return ERROR_OBJECT_EXISTS;
	}
    }

    /* One or more FileExtBlocks are needed */

    if(!~file->deb_Extension)
    {
	/* New FileExtBlock */
	block2 = AllocBlock(ALLOC_INFO);
	if(block2 == -1)
	    return ERROR_DISK_FULL;
	FlashWriteLong(&file->deb_Extension, block2);
    }
    blockoffset -= debsize;
    feb = (struct FileExtBlock *)GetBlock(file->deb_Extension);

    while(blockoffset >= febsize)
    {
	if(!~feb->feb_Extension)
	{
	    block2 = AllocBlock(ALLOC_INFO);
	    if(block2 == -1)
		return ERROR_DISK_FULL;
	    FlashWriteLong(&feb->feb_Extension, block2);
	}
	blockoffset -= febsize;
	feb = (struct FileExtBlock *)GetBlock(feb->feb_Extension);
    }

    if(!~feb->feb_Data[blockoffset])
    {
	FlashWriteLong(&feb->feb_Data[blockoffset], newblk);
	return 0;
    }
    else
    {
	/* This block already written */
	return ERROR_OBJECT_EXISTS;
    }
}

LONG TellFileBlock(struct DirEntryBlock *file, ULONG blockoffset)
{
    ULONG debsize = (bb->bb_BlockSize/4)-60,
	  febsize = (bb->bb_BlockSize/4)-2;
    struct FileExtBlock *feb;

    if(blockoffset < debsize)
    {
	/* There is a pointer in the DirEntryBlock */
	if(!~file->deb_HashTable[blockoffset])
	{
	    return -1;
	}
	return (LONG)file->deb_HashTable[blockoffset];
    }

    /* One or more FileExtBlocks are needed */

    if(!~file->deb_Extension)
    {
	/* No FileExtBlock */
	return -1;
    }
    blockoffset -= debsize;
    feb = (struct FileExtBlock *)GetBlock(file->deb_Extension);

    while(blockoffset >= febsize)
    {
	if(!~feb->feb_Extension)
	{
	    return -1;
	}
	blockoffset -= febsize;
	feb = (struct FileExtBlock *)GetBlock(feb->feb_Extension);
    }

    if(!~feb->feb_Data[blockoffset])
    {
	return -1;
    }
    return (LONG)feb->feb_Data[blockoffset];
}

struct DirEntryBlock *CopyBlock(struct DirEntryBlock *block)
{
    struct DirEntryBlock *deb =
	(struct DirEntryBlock *)AllocMem(bb->bb_BlockSize, 0);

    if(!deb)
	return 0;
    memcpy(deb, block, sizeof(struct DirEntryBlock));
    return deb;
}

void FreeCopyBlock(struct DirEntryBlock *block)
{
    FreeMem((char *)block, bb->bb_BlockSize);
}


LONG ActionFormat(struct DosPacket *packet)
{
    const ULONG blocksize = 0x100;
    const ULONG lowblock = 0x840000/blocksize;
    const ULONG hiblock = (0xc00000/blocksize)-1;
    const ULONG rootblock = (lowblock+hiblock)/2;
    /* RAM copy of the root block/bit map for initialization */
    struct DirEntryBlock *root =
	(struct DirEntryBlock *)AllocMem(blocksize, MEMF_CLEAR);
    struct Task *task = (struct Task *)
	packet->dp_Message.mn_ReplyPort->mp_SigTask;
    ULONG poslong,posbit, hashsize = (blocksize/4)-60;
    ULONG bitmapblocks = (hiblock-lowblock)/8/blocksize + 1;
    struct BitMapBlock *bitmap =
	(struct BitMapBlock *)AllocMem(bitmapblocks*blocksize, 0);
    int i;

    if(!root || !bitmap)
    {
	if(root)
	    FreeMem((char *)root, blocksize);
	if(bitmap)
	    FreeMem((char *)bitmap, bitmapblocks*blocksize);

	packet->dp_Result2 = ERROR_NO_FREE_STORE;
	return DOS_FALSE;
    }

    /* Well, don't let anyone to format someone else's disk.. */
    if(bb->bb_DiskType == ID_DOS_DISK &&
	(GetBlock(bb->bb_RootBlock)->deb_Owner>>16) != task->tc_UserID)
    {
	FreeMem((char *)root, blocksize);
	FreeMem((char *)bitmap, bitmapblocks*blocksize);

	packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	return DOS_FALSE;
    }

    /* Currently hard-coded for ROM */
    FlashErase(); /* Init Flash to all ones */

    FlashWriteLong(&bb->bb_DiskType, ID_DOS_DISK);

    FlashWriteLong(&bb->bb_LowBlock, lowblock);
    FlashWriteLong(&bb->bb_HiBlock,  hiblock);
    FlashWriteLong(&bb->bb_BlockSize,blocksize);
    FlashWriteLong(&bb->bb_RootBlock,rootblock);
    FlashWriteLong(&bb->bb_BitMapBlock, rootblock+1);
    FlashWriteLong(&bb->bb_BitMapBlocks,bitmapblocks);

    root->deb_Type = 2;
    DateStamp(&root->deb_ChangeDate);
    strcpy(root->deb_FileName, "ROM");
    root->deb_Owner = (task->tc_UserID<<16) | task->tc_GroupID;
    root->deb_Protect = 0x111 * (FIBF_READ | FIBF_WRITE | FIBF_EXECUTE);
    memset(&root->deb_HashTable[0], 0xff, 4*hashsize);

    WriteBlock(rootblock, (ULONG *)root, (int)blocksize);

    memset(bitmap, 0xff, bitmapblocks*blocksize);
    /* Reserve BootBlock, RootBlock & BitMapBlock(s) */
    bitmap->bmb_BitMap[0] &= ~1U;
    bitmap->bmb_BitMap[(rootblock-lowblock)/32] &= ~(1U<<((rootblock-lowblock)%32));
    for(i=0;i<bitmapblocks;i++)
    {
	bitmap->bmb_BitMap[(rootblock+1+i-lowblock)/32] &=
		~(1U<<((rootblock+1+i-lowblock)%32));
    }

    poslong = (bb->bb_HiBlock-bb->bb_LowBlock)/32;
    bitmap->bmb_BitMap[poslong] = 0;
    posbit = 0;
    while(bb->bb_LowBlock + poslong*32 + posbit <= bb->bb_HiBlock)
    {
	bitmap->bmb_BitMap[poslong] |= (1U<<posbit);
	posbit++;
    }

    WriteBlock(rootblock+1, (ULONG *)bitmap, (int)bitmapblocks*(int)blocksize);

    FreeMem((char *)root, blocksize);
    FreeMem((char *)bitmap, bitmapblocks*blocksize);
    return DOS_TRUE;
}


struct DirEntryBlock *LocateEntry(struct DirEntryBlock *base, char *name, int parent)
{
    char namecomp[108];
    int i = 0;
    struct DirEntryBlock *found = NULL;
    UWORD hash, hashsize = (bb->bb_BlockSize/4)-60;

    while(1)
    {
	if(!*name)
	{
	    /* End of the name */
	    namecomp[i] = '\0';
	    /* Looking for a parent, "" or "." */
	    if(parent || !i || !strcmp(namecomp, "."))
		return base;

	    if(!strcmp(namecomp, ".."))
		return GetBlock(base->deb_ParentDir);

	    if(!base || base->deb_Type<0)
		return NULL;

	    hash = CalcHashNoCase(namecomp) % hashsize;

	    /* GetBlock() returns NULL also if the parameter is 0xffffffff */
	    found = GetBlock(base->deb_HashTable[hash]);
	    while(found)
	    {
		if(found->deb_Type && !strcmp(found->deb_FileName, namecomp))
		{
		    return found;
		}
		found = GetBlock(found->deb_HashChain);
	    }
	    return NULL;
	}

	if(*name==':')
	{
	    i = 0;
	    /*
	       Handler name can be for this handler, which means we
	       got a lock to the root directory (or NULL). Otherwise
	       the handler name can be an assign and we got a lock
	       to a directory.
	       */
	}
	else if(*name=='/')
	{
	    namecomp[i] = '\0';
	    if(!strcmp(namecomp, "."))
	    {
		/* ./ goes nowhere */
		/* base = base; */
	    }
	    else if(!i || !strcmp(namecomp, ".."))
	    {
		/* "../" or "/" makes us go to the parent directory */
		if(!base->deb_ParentDir && !~base->deb_ParentDir)
		    return NULL;
		base = GetBlock(base->deb_ParentDir);
	    }
	    else
	    {
		hash = CalcHashNoCase(namecomp) % hashsize;
		found = GetBlock(base->deb_HashTable[hash]);
		while(found)
		{
		    if(found->deb_Type && !strcmp(found->deb_FileName, namecomp))
		    {
			break;
		    }
		    found = GetBlock(found->deb_HashChain);
		}

		/* entry not found or not a directory (or deleted) */
		if(!found || (found->deb_Type <= 0))
		    return NULL;
		base = found;
	    }
	    i = 0;
	}
	else
	{
	    if(i<106)
		namecomp[i++] = *name;
	}
	name++;
    }
}


struct FileLock *FindLock(ULONG key)
{
    struct FileLock *lock = 
	(struct FileLock *)dol->dol_misc.dol_volume.dol_LockList.lh_Head;

    while(lock->fl_Node.lmn_Next)
    {
	if(lock->fl_Key == key)
	    return lock;
	lock = (struct FileLock *)lock->fl_Node.lmn_Next;
    }
    return NULL;
}

struct FileLock *LockEntry(LONG *err, ULONG key, LONG mode)
{
    struct FileLock *lock = FindLock(key), *newlock;

    if((mode == EXCLUSIVE_LOCK) && lock)
    {
	*err = ERROR_OBJECT_IN_USE;
	return NULL;
    }

    if((mode == SHARED_LOCK) && lock && lock->fl_Access!=SHARED_LOCK)
    {
	*err = ERROR_OBJECT_IN_USE;
	return NULL;
    }

    newlock = (struct FileLock *)AllocMem(sizeof(struct FileLock), MEMF_CLEAR);
    if(!newlock)
    {
	*err = ERROR_NO_FREE_STORE;
	return NULL;
    }

    newlock->fl_Key = key;
    newlock->fl_Access = mode;
    newlock->fl_Task = myport;
    AddTail(&dol->dol_misc.dol_volume.dol_LockList, (struct Node *)newlock);
    return newlock;
}

void UnLockEntry(struct FileLock *lock)
{
    Remove((struct Node *)lock);
    FreeMem(lock, sizeof(struct FileLock));
}

LONG ActionLock(struct DosPacket *packet)
{
    struct DirEntryBlock *dirent =
	LocateEntry(GetBaseDir(packet->dp_Arg1), (char *)packet->dp_Arg2, 0);
    ULONG key = (ULONG)((ULONG)dirent/bb->bb_BlockSize);

    if(!dirent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    return (LONG)LockEntry(&packet->dp_Result2, key, (LONG)packet->dp_Arg3);
}

LONG ActionUnLock(struct DosPacket *packet)
{
    UnLockEntry((struct FileLock *)packet->dp_Arg1);
    return DOS_TRUE;
}


/*
   fh_Buf = (UBYTE *)buffer,
   fh_Pos = (ULONG)bufind,
   fh_End = file size in blocks this far
   fh_Arg1= (DirEntryBlock *)
*/

LONG ActionOpen(struct DosPacket *packet)
{
    struct DirEntryBlock *deb = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 0);
    struct FileHandle *fh;
    struct DirEntryBlock *parent = LocateEntry(GetBaseDir(packet->dp_Arg1),
					  (char *)packet->dp_Arg2, 1);
    struct Task *task = packet->dp_Message.mn_ReplyPort->mp_SigTask;
    ULONG Owner = (task->tc_UserID<<16) | task->tc_GroupID;
    char *name = FilePart((char *)packet->dp_Arg2);
    LONG block;

    if(!parent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!strcmp(name, "..") || !strcmp(name, "."))
    {
	packet->dp_Result2 = ERROR_OBJECT_EXISTS;
	return DOS_FALSE;
    }
    if(deb && deb->deb_Type > 0)
    {
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }

    fh = (struct FileHandle *)AllocMem(sizeof(struct FileHandle), MEMF_CLEAR);
    if(!fh)
    {
	packet->dp_Result2 = ERROR_NO_FREE_STORE;
	return DOS_FALSE;
    }

    switch(packet->dp_Arg3) /* Open mode */
    {
    case MODE_OLDFILE:
	if(!deb)
	{
	    packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	if(!GetAccess(packet, deb, FIBF_READ))
	{
	    packet->dp_Result2 = ERROR_READ_PROTECTED;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	fh->fh_Port = myport;
	fh->fh_Arg1 = (ULONG)deb;
	fh->fh_Buf  = 0; /* does not point to a buffer, also: Read mode */
	fh->fh_Pos  = 0;
	fh->fh_End  = 0; /* No blocks read */
	fh->fh_Arg2 = (ULONG)LockEntry(&packet->dp_Result2,
				       (((ULONG)deb)/bb->bb_BlockSize),
				       (LONG)SHARED_LOCK);
	return (LONG)fh;
	break;
    case MODE_NEWFILE:
	if(deb)
	{
	    if(!GetAccess(packet, deb, FIBF_DELETE))
	    {
		packet->dp_Result2 = ERROR_DELETE_PROTECTED;
		FreeMem(fh, sizeof(struct FileHandle));
		return DOS_FALSE;
	    }
	    if((packet->dp_Result2 =
		DeleteDirEntry(deb, (ULONG)deb/bb->bb_BlockSize)))
	    {
		FreeMem(fh, sizeof(struct FileHandle));
		return DOS_FALSE;
	    }
	}
	if(!GetAccess(packet, parent, FIBF_WRITE))
	{
	    packet->dp_Result2 = ERROR_PARENT_PROTECTED;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	deb = (struct DirEntryBlock *)AllocMem(bb->bb_BlockSize, MEMF_CLEAR);
	if(!deb)
	{
	    packet->dp_Result2 = ERROR_NO_FREE_STORE;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	block = AllocBlock(ALLOC_INFO);
	if(block == -1)
	{
	    FreeMem(deb, bb->bb_BlockSize);
	    packet->dp_Result2 = ERROR_DISK_FULL;
	    FreeMem(fh, sizeof(struct FileHandle));
	    return DOS_FALSE;
	}
	InitDirEntry(deb, -1, parent, name, Owner);
	WriteBlock(block, (ULONG *)deb, (int)bb->bb_BlockSize);
	fh->fh_Port = myport;
	fh->fh_Arg1 = (ULONG)GetBlock(block);
	fh->fh_Buf  = (ULONG)AllocMem(bb->bb_BlockSize, 0);
	fh->fh_Pos  = 0;
	fh->fh_End  = 0; /* No blocks written */
	fh->fh_Arg2 = (ULONG)LockEntry(&packet->dp_Result2,
				       (((ULONG)deb)/bb->bb_BlockSize),
				       (LONG)EXCLUSIVE_LOCK);

	AddDirEntry(parent, deb->deb_FileName, block);
	return (LONG)fh;
	break;
    default:
        packet->dp_Result2 = ERROR_NOT_IMPLEMENTED;
	FreeMem(fh, sizeof(struct FileHandle));
	return DOS_FALSE;
    }
    FreeMem(fh, sizeof(struct FileHandle));
    return DOS_FALSE;
}

LONG ActionClose(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    struct DirEntryBlock *deb = (struct DirEntryBlock *)fh->fh_Arg1;

    if(fh->fh_Buf)
    {
	if(fh->fh_Pos) /* Anything to flush ? */
	{
	    /* 'Clear' unused bytes */
	    memset((char *)fh->fh_Buf+fh->fh_Pos, 0xff, bb->bb_BlockSize-fh->fh_Pos);
	    if((packet->dp_Result2 =
		WriteFileBlock(deb, fh->fh_End, (ULONG *)fh->fh_Buf)))
	    {
		FreeMem((char *)fh->fh_Buf, bb->bb_BlockSize);
		FlashWriteLong(&deb->deb_ByteSize,
			       fh->fh_End*bb->bb_BlockSize+fh->fh_Pos);
		return DOS_FALSE;
	    }
	}
	FreeMem((char *)fh->fh_Buf, bb->bb_BlockSize);
	FlashWriteLong(&deb->deb_ByteSize,
		       fh->fh_End*bb->bb_BlockSize+fh->fh_Pos);
    }
    /* UnLock */
    if(fh->fh_Arg2)
	UnLockEntry((struct FileLock *)fh->fh_Arg2);
    FreeMem(fh, sizeof(struct FileHandle));
    return DOS_TRUE;
}


LONG ActionRead(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    struct DirEntryBlock *deb = (struct DirEntryBlock *)fh->fh_Arg1;
    char *buffer = (char *)packet->dp_Arg2;
    LONG size = packet->dp_Arg3, read = 0, left, block;
    LONG blockoffset, byteoffset;

    while(size && (fh->fh_Pos < deb->deb_ByteSize))
    {
	blockoffset = fh->fh_Pos/bb->bb_BlockSize,
	byteoffset = fh->fh_Pos%bb->bb_BlockSize;

	left = bb->bb_BlockSize - byteoffset;
	if(size < left)
	    left = size;
	if(fh->fh_Pos + left > deb->deb_ByteSize)
	{
	    left = deb->deb_ByteSize - fh->fh_Pos;
	}
	block = TellFileBlock(deb, blockoffset);
	if(block == -1)
	    return read;
	memcpy(buffer, (char *)GetBlock(block)+byteoffset, left);
	fh->fh_Pos += left;
	read += left;
	size -= left;
	buffer += left;
    }
    packet->dp_Result2 = 0;
    return read;
}

LONG ActionWrite(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    struct DirEntryBlock *deb = (struct DirEntryBlock *)fh->fh_Arg1;
    char *buffer = (char *)packet->dp_Arg2;
    LONG size = (LONG)packet->dp_Arg3, left;

    while(size)
    {
	left = bb->bb_BlockSize - fh->fh_Pos;
	if(size < left)
	    left = size;
	memcpy((char *)fh->fh_Buf+fh->fh_Pos, buffer, left);
	size -= left;
	buffer += left;
	fh->fh_Pos += left;
	if(fh->fh_Pos == bb->bb_BlockSize)
	{
	    fh->fh_Pos = 0;
	    /* Buffer full */
	    if((packet->dp_Result2 =
		WriteFileBlock(deb, fh->fh_End, (ULONG *)fh->fh_Buf)))
	    {
		return DOS_FALSE;
	    }
	    fh->fh_End++;
	}
    }
    return (LONG)packet->dp_Arg3-size;
}

LONG ActionSeek(struct DosPacket *packet)
{
    struct FileHandle *fh = (struct FileHandle *)packet->dp_Arg1;
    struct DirEntryBlock *deb = (struct DirEntryBlock *)fh->fh_Arg1;
    LONG offset = (LONG)packet->dp_Arg2;
    LONG mode   = (LONG)packet->dp_Arg3;
    LONG oldloc, loc;

    /* seeks only possible while reading */
    if(fh->fh_Buf)
    {
	packet->dp_Result2 = ERROR_NOT_IMPLEMENTED;
	return DOS_FALSE;
    }

    if(mode==OFFSET_END)
    {
	loc = deb->deb_ByteSize + offset;
    }
    else if(mode==OFFSET_BEGINNING)
    {
	loc = offset;
    }
    else if(mode==OFFSET_CURRENT)
    {
	loc = fh->fh_Pos + offset;
    }
    else
    {
	packet->dp_Result2 = ERROR_SEEK_ERROR;
	return DOS_FALSE;
    }
    if(loc < 0 || loc > deb->deb_ByteSize)
    {
	packet->dp_Result2 = ERROR_SEEK_ERROR;
	return DOS_FALSE;
    }
    oldloc = fh->fh_Pos;
    fh->fh_Pos = loc;
    packet->dp_Result2 = (LONG)oldloc;
    return DOS_TRUE;
}


LONG ActionCreateDir(struct DosPacket *packet)
{
    struct DirEntryBlock *parent = LocateEntry(GetBaseDir(packet->dp_Arg1),
						(char *)packet->dp_Arg2, 1);
    struct DirEntryBlock *deb;
    char *name = FilePart((char *)packet->dp_Arg2);
    struct FileLock *fl;
    struct Task *task = packet->dp_Message.mn_ReplyPort->mp_SigTask;
    ULONG Owner = (task->tc_UserID<<16) | task->tc_GroupID, block;

    if(!*name || !parent)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(strlen(name)>106)
    {
	packet->dp_Result2 = ERROR_INVALID_COMPONENT_NAME;
	return DOS_FALSE;
    }
    if(!strcmp(name, "..") || !strcmp(name, "."))
    {
	packet->dp_Result2 = ERROR_OBJECT_EXISTS;
	return DOS_FALSE;
    }

    deb = LocateEntry(parent, name, 0);
    if(deb)
    {
	packet->dp_Result2 = ERROR_OBJECT_EXISTS;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, parent, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_PARENT_PROTECTED;
	return DOS_FALSE;
    }

    deb = (struct DirEntryBlock *)AllocMem(bb->bb_BlockSize, MEMF_CLEAR);
    if(!deb)
    {
	packet->dp_Result2 = ERROR_NO_FREE_STORE;
	return DOS_FALSE;
    }
    block = AllocBlock(ALLOC_INFO);
    if(block == -1)
    {
	FreeMem(deb, bb->bb_BlockSize);
	packet->dp_Result2 = ERROR_DISK_FULL;
	return DOS_FALSE;
    }

    /* OK, we have a block. Fill in the new direntry, write it and
       add it to the hash chain of the parent */

    if(fl = LockEntry(&packet->dp_Result2, block, SHARED_LOCK)) /* Should always succeed */
    {
	InitDirEntry(deb, 1, parent, name, Owner);

	AddDirEntry(parent, name, block);
	WriteBlock(block, (ULONG *)deb, (int)bb->bb_BlockSize);

	FreeMem(deb, bb->bb_BlockSize);
	return (LONG)fl;
    }
    /* If it fails, leave the block allocated, only release memory */
    FreeMem(deb, bb->bb_BlockSize);
    return DOS_FALSE;
}

LONG ActionDelete(struct DosPacket *packet)
{
    struct DirEntryBlock *deb = LocateEntry(GetBaseDir(packet->dp_Arg1),
					(char *)packet->dp_Arg2, 0);
    if(!deb)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(!GetAccess(packet, deb, FIBF_DELETE))
    {
	packet->dp_Result2 = ERROR_DELETE_PROTECTED;
	return DOS_FALSE;
    }
    if((packet->dp_Result2 = DeleteDirEntry(deb, (ULONG)deb/bb->bb_BlockSize)))
	return DOS_FALSE;
    return DOS_TRUE;
}

LONG ActionExamine(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct FileInfoBlock *fib = (struct FileInfoBlock *)packet->dp_Arg2;
    struct DirEntryBlock *deb = GetBlock(fl->fl_Key);

    InitFileInfoBlock(fib, deb, fl->fl_Key);

    fib->fib_Reserved[7] = -1; /* Hash number 0 next */
    return DOS_TRUE;
}

LONG ActionExNext(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct FileInfoBlock *fib = (struct FileInfoBlock *)packet->dp_Arg2;
    struct DirEntryBlock *dir = GetBlock(fl->fl_Key);
    struct DirEntryBlock *deb = GetBlock(fib->fib_DiskKey);
    ULONG hashsize = (bb->bb_BlockSize/4)-60;

    if(dir->deb_Type < 0)
    {
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }

    if(fib->fib_Reserved[7]==-1 || fib->fib_DiskKey==fl->fl_Key)
    {
	fib->fib_Reserved[7] = -1;
	deb = NULL;
    }
    /* If we have anything, get the next entry, if any */
    if(deb)
    {
	deb = GetBlock(deb->deb_HashChain);

	/* Skip invalid entries */
	while(deb && !deb->deb_Type)
	    deb = GetBlock(deb->deb_HashChain);
    }

    /* If we got to the end of a hash chain or we have an empty chain */
    while(!deb && (++fib->fib_Reserved[7]<hashsize))
    {
	/* Get the first file in the directory */
	deb = GetBlock(dir->deb_HashTable[fib->fib_Reserved[7]]);

	/* Skip invalid entries */
	while(deb && !deb->deb_Type)
	    deb = GetBlock(deb->deb_HashChain);
    }
    if(deb)
    {
	InitFileInfoBlock(fib, deb, (ULONG)deb/bb->bb_BlockSize);
	return DOS_TRUE;
    }
    packet->dp_Result2 = ERROR_NO_MORE_ENTRIES;
    return DOS_FALSE;
}

LONG ActionExamineEnd(struct DosPacket *packet)
{
#if 0
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct FileInfoBlock *fib = (struct FileInfoBlock *)packet->dp_Arg2;

    /* No buffering done yet, so no cleanup to do either */
#endif
    return DOS_TRUE;
}

LONG ActionCurrentDir(struct DosPacket *packet)
{
    struct FileLock *fl = (struct FileLock *)packet->dp_Arg1;
    struct DirEntryBlock *dir = GetBlock(fl->fl_Key);

    if(dir->deb_Type < 0)
    {
	packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	return DOS_FALSE;
    }
    return (LONG)fl;
}

LONG ActionSameLock(struct DosPacket *packet)
{
    struct FileLock *flA = (struct FileLock *)packet->dp_Arg1;
    struct FileLock *flB = (struct FileLock *)packet->dp_Arg2;

    if(flA->fl_Key == flB->fl_Key)
	return DOS_TRUE;
    return DOS_FALSE;
}

LONG ActionParentDir(struct DosPacket *packet)
{
    struct FileLock *fl;
    struct DirEntryBlock *deb = GetBaseDir(packet->dp_Arg1);

    packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
    if(deb == GetBlock(bb->bb_RootBlock))
    {
	return DOS_FALSE;
    }
    if(deb->deb_ParentDir &&
       (fl = LockEntry(&packet->dp_Result2, deb->deb_ParentDir, ACCESS_READ)))
    {
	return (LONG)fl;
    }
    return DOS_FALSE;
}

LONG ActionSetComment(struct DosPacket *packet)
{
    struct DirEntryBlock *deb = LocateEntry(GetBaseDir(packet->dp_Arg1),
					    (char *)packet->dp_Arg2, 0);
    char temp[80];
    int i;
    struct DirEntryBlock *debnew;
    LONG block;
    struct FileLock *lock;

    if(!deb)
    {
        packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
        return DOS_FALSE;
    }
    if(!GetAccess(packet, deb, FIBF_WRITE))
    {
        packet->dp_Result2 = ERROR_WRITE_PROTECTED;
        return DOS_FALSE;
    }
    if(strlen((char *)packet->dp_Arg3)>79)
    {
        packet->dp_Result2 = ERROR_COMMENT_TOO_BIG;
        return DOS_FALSE;
    }
    if(deb->deb_Comment[0] == 0xff)
    {
	memset(temp, 0xff, 80);
	strcpy(temp, (char *)packet->dp_Arg3);
	for(i=0;i<20;i++)
	{
	    FlashWriteLong((ULONG *)(deb->deb_Comment+4*i), *(((ULONG *)temp)+i));
	}
	return DOS_TRUE;
    }

    /********************************************/
    /* 1) allocate new direntry                 */
    /* 2) copy data from the old direntry       */
    /* 3) change the data that is to be changed */
    /* 4) delete old direntry (mark deleted)    */
    /* 5) add new entry                         */
    /********************************************/
    /* Check if the entry can be reallocated */
    lock = FindLock((ULONG)deb/bb->bb_BlockSize);
    if(lock)
    {
	packet->dp_Result2 = ERROR_OBJECT_IN_USE;
	return DOS_FALSE;
    }
    /* Allocate a block for the new direntry */
    block = AllocBlock(ALLOC_INFO);
    if(block == -1)
    {
	packet->dp_Result2 = ERROR_DISK_FULL;
	return DOS_FALSE;
    }
    debnew = CopyBlock(deb);
    if((packet->dp_Result2 = DeleteDirEntry(deb, (ULONG)deb/bb->bb_BlockSize)))
    {
	/* should not happen! */
	FreeCopyBlock(debnew);
	return DOS_FALSE;
    }
    debnew->deb_HashChain = ~0;
    memset(debnew->deb_Comment, 0xff, 80);
    strcpy(debnew->deb_Comment, (char *)packet->dp_Arg3);
    WriteBlock(block, (ULONG *)debnew, (int)bb->bb_BlockSize);
    AddDirEntry(GetBlock(debnew->deb_ParentDir), debnew->deb_FileName, block);
    FreeCopyBlock(debnew);
    return DOS_TRUE;
}


LONG ActionSetOwner(struct DosPacket *packet)
{
    struct DirEntryBlock *deb = LocateEntry(GetBaseDir(packet->dp_Arg1),
					    (char *)packet->dp_Arg2, 0);
    struct DirEntryBlock *debnew;
    LONG block;
    struct FileLock *lock;

    if(!deb)
    {
        packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
        return DOS_FALSE;
    }
    if(!IsOwner(packet, deb))
    {
        packet->dp_Result2 = ERROR_NOT_OWNER;
        return DOS_FALSE;
    }
    if((packet->dp_Arg3 & deb->deb_Owner)==packet->dp_Arg3)
    {
	FlashWriteLong(&deb->deb_Owner, packet->dp_Arg3);
	return DOS_TRUE;
    }

    /********************************************/
    /* 1) allocate new direntry                 */
    /* 2) copy data from the old direntry       */
    /* 3) change the data that is to be changed */
    /* 4) delete old direntry (mark deleted)    */
    /* 5) add new entry                         */
    /********************************************/
    /* Check if the entry can be reallocated */
    lock = FindLock((ULONG)deb/bb->bb_BlockSize);
    if(lock)
    {
	packet->dp_Result2 = ERROR_OBJECT_IN_USE;
	return DOS_FALSE;
    }
    /* Allocate a block for the new direntry */
    block = AllocBlock(ALLOC_INFO);
    if(block == -1)
    {
	packet->dp_Result2 = ERROR_DISK_FULL;
	return DOS_FALSE;
    }
    debnew = CopyBlock(deb);
    if((packet->dp_Result2 = DeleteDirEntry(deb, (ULONG)deb/bb->bb_BlockSize)))
    {
	/* should not happen! */
	FreeCopyBlock(debnew);
	return DOS_FALSE;
    }
    debnew->deb_HashChain = ~0;
    debnew->deb_Owner = packet->dp_Arg3;
    WriteBlock(block, (ULONG *)debnew, (int)bb->bb_BlockSize);
    AddDirEntry(GetBlock(debnew->deb_ParentDir), debnew->deb_FileName, block);
    FreeCopyBlock(debnew);
    return DOS_TRUE;
}

LONG ActionSetProtection(struct DosPacket *packet)
{
    struct DirEntryBlock *deb = LocateEntry(GetBaseDir(packet->dp_Arg1),
					    (char *)packet->dp_Arg2, 0);
    struct DirEntryBlock *debnew;
    LONG block;
    struct FileLock *lock;

    if(!deb)
    {
        packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
        return DOS_FALSE;
    }
    if(!IsOwner(packet, deb))
    {
        packet->dp_Result2 = ERROR_NOT_OWNER;
        return DOS_FALSE;
    }
    if((packet->dp_Arg3 & deb->deb_Protect)==packet->dp_Arg3)
    {
	FlashWriteLong(&deb->deb_Protect, packet->dp_Arg3);
	return DOS_TRUE;
    }

    /********************************************/
    /* 1) allocate new direntry                 */
    /* 2) copy data from the old direntry       */
    /* 3) change the data that is to be changed */
    /* 4) delete old direntry (mark deleted)    */
    /* 5) add new entry                         */
    /********************************************/
    /* Check if the entry can be reallocated */
    lock = FindLock((ULONG)deb/bb->bb_BlockSize);
    if(lock)
    {
	packet->dp_Result2 = ERROR_OBJECT_IN_USE;
	return DOS_FALSE;
    }
    /* Allocate a block for the new direntry */
    block = AllocBlock(ALLOC_INFO);
    if(block == -1)
    {
	packet->dp_Result2 = ERROR_DISK_FULL;
	return DOS_FALSE;
    }
    debnew = CopyBlock(deb);
    if((packet->dp_Result2 = DeleteDirEntry(deb, (ULONG)deb/bb->bb_BlockSize)))
    {
	/* should not happen! */
	FreeCopyBlock(debnew);
	return DOS_FALSE;
    }
    debnew->deb_HashChain = ~0;
    debnew->deb_Protect = packet->dp_Arg3;
    WriteBlock(block, (ULONG *)debnew, (int)bb->bb_BlockSize);
    AddDirEntry(GetBlock(debnew->deb_ParentDir), debnew->deb_FileName, block);
    FreeCopyBlock(debnew);
    return DOS_TRUE;
}

LONG ActionRename(struct DosPacket *packet)
{
    LONG block;
    struct DirEntryBlock *debold = LocateEntry(GetBaseDir(packet->dp_Arg1),
					       (char *)packet->dp_Arg2, 0);
    struct DirEntryBlock *destdir= LocateEntry(GetBaseDir(packet->dp_Arg3),
					       (char *)packet->dp_Arg4, 1);
    struct DirEntryBlock *debnew;
    struct FileLock *lock;

    if(!debold || !destdir)
    {
	packet->dp_Result2 = ERROR_OBJECT_NOT_FOUND;
	return DOS_FALSE;
    }
    if(debold == GetBlock(bb->bb_RootBlock) ||
       !GetAccess(packet, debold, FIBF_WRITE))
    {
        packet->dp_Result2 = ERROR_WRITE_PROTECTED;
        return DOS_FALSE;
    }
    debnew = LocateEntry(destdir, FilePart((char *)packet->dp_Arg4), 0);
    if(debold == debnew)
    {
	return DOS_TRUE;
    }

    if(debnew)
    {
	if(debnew->deb_Type > 0)
	{
	    /* TODO: put the file IN the directory ?? */
	    packet->dp_Result2 = ERROR_OBJECT_WRONG_TYPE;
	    return DOS_FALSE;
	}
	if(!GetAccess(packet, debnew, FIBF_WRITE))
	{
	    packet->dp_Result2 = ERROR_WRITE_PROTECTED;
	    return DOS_FALSE;
	}
	if((packet->dp_Result2 = DeleteDirEntry(debnew, (ULONG)debnew/bb->bb_BlockSize)))
	{
	    return DOS_FALSE;
	}
    }

    /* If there was an old file, it is now deleted */
    if(!GetAccess(packet, destdir, FIBF_WRITE))
    {
	packet->dp_Result2 = ERROR_PARENT_PROTECTED;
	return DOS_FALSE;
    }

    /********************************************/
    /* 1) allocate new direntry                 */
    /* 2) copy data from the old direntry       */
    /* 3) change the data that is to be changed */
    /* 4) delete old direntry (mark deleted)    */
    /* 5) add new entry                         */
    /********************************************/
    /* Check if the entry can be renamed */
    lock = FindLock((ULONG)debold/bb->bb_BlockSize);
    if(lock)
    {
	packet->dp_Result2 = ERROR_OBJECT_IN_USE;
	return DOS_FALSE;
    }

    /* Allocate a block for the new direntry */
    block = AllocBlock(ALLOC_INFO);
    if(block == -1)
    {
	packet->dp_Result2 = ERROR_DISK_FULL;
	return DOS_FALSE;
    }

    debnew = CopyBlock(debold);
    if((packet->dp_Result2 = DeleteDirEntry(debold, (ULONG)debold/bb->bb_BlockSize)))
    {
	/* This should not happen.. */
	FreeCopyBlock(debnew);
	return DOS_FALSE;
    }

    strncpy(debnew->deb_FileName, FilePart((char *)packet->dp_Arg4), 107);
    debnew->deb_FileName[107] = '\0';
    debnew->deb_HashChain = ~0;
    debnew->deb_ParentDir = ((ULONG)destdir)/bb->bb_BlockSize;
    WriteBlock(block, (ULONG *)debnew, (int)bb->bb_BlockSize);
    AddDirEntry(destdir, debnew->deb_FileName, block);
    FreeCopyBlock(debnew);
    return DOS_TRUE;
}
