#include <stdio.h>
#include <stdlib.h>

/************************************************************************\
* Xlink version 1.1      1994 Pasi 'Albert' Ojala                        *
*                                                                        *
* What is Xlink for ?                                                    *
*     Xlink takes a a.out-format incremental link (or object file)       *
*     and makes it ready for relocatable loading.                        *
* Why Xlink ?                                                            *
*     Normal U*ix ld produces executables linked in a fixed address,     *
*     because in U*ix each program has its own a virtual address space.  *
*     As a result, relocation information is lost when linking.          *
*     The linker can be used for 'incremental linking', where the        *
*     relocation information is retained, BUT it won't complain if       *
*     external symbols remain unresolved. Also, common uninitialized     *
*     data regions are not allocated from BSS-section until the final    *
*     phase, thus there remains unresolved external references in the    *
*     incremental linking.                                               *
* What does Xlink do exactly ?                                           *
*     Xlink reads in a a.out-format incremental link (or object file),   *
*     goes through the relocation information and allocates all          *
*     uninitialized common data regions while also resolving any         *
*     references to them. (normally done by "ld" during the final        *
*     linking phase). In addition, offset-values for data-relocations    *
*     are changed to relative to the data-segment start instead of the   *
*     text-segment start. This is because BOAR exec will allocate code,  *
*     data and BSS areas separately, unlike U*ix, which has code, data   *
*     and BSS consecutively located in memory.                           *
*     Also, Xlink changes the a_toolversion to 68.                       *
*     Symbol information and stringsection is removed.                   *
*                                                                        *
* What else can you do with Xlink ?                                      *
*  1) You can generate raw files from executables (Xlink -r).            *
*     This is needed when you like to transfer the program into E(E)PROM.*
*     "Xlink -r" generates three files. A file with the suffix .raw is   *
*     intended for byte-wide data buses or for eprommer programs that    *
*     know how to divide the data into two EPROMs. Files with suffixes   *
*     .even and .odd hold the even and odd bytes respectively. These     *
*     files are needed when your eprommer program can't divide raw files *
*     into two EPROMs.                                                   *
*     REMEMBER: If the starting address is not zero, divide it by two    *
*               when you are making the eproms for a 16-bit system with  *
*               two EPROMs.                                              *
*  2) You can make linkable object files from raw data files (Xlink -d). *
*     When you use some program to generate data used in the progam,     *
*     it is not usually very convenient to import the data to the        *
*     source code each time the data changes, especially if there is     *
*     lots of it. However, it is very convenient to use Xlink -d to      *
*     generate a linkable file out of the data. You can then refer to    *
*     the data with two symbols, _<filename> and _<filename>_size.       *
*     (Prepend a '_' for assembly.) The first symbol defines the ADDRESS *
*     of the data, the second one tells the length of the data.          *
*     The data will be located in the text segment.                      *
*     Example:                                                           *
*             >  ./xlink -d -o xildata.o ClockXilinxData                 *
*             test.c:                                                    *
*             extern const UBYTE _ClockXilinxData[];                     *
*             extern const ULONG _ClockXilinxData_size;                  *
*                                                                        *
\************************************************************************/


#if 0
/* Our system does not have proper <stdio.h> ! */
int printf(char *, ...);
int fprintf(FILE *, char *, ...);
int fread(char *, int, int, FILE *);
int fwrite(char *,int,int,FILE *);
int fseek(FILE *, int, int);
void fclose(FILE *);
#endif

/* Some shorthands */
#define UWORD unsigned short
#define ULONG unsigned long
#define UBYTE unsigned char

/* a.out header - a_toolversion includes Static-flag */
struct exec
{
    UBYTE   a_toolversion;  /* version of toolset used to create */
    UBYTE   a_machtype;	/* machine type */
    UWORD   a_magic;	/* magic number */
    
    ULONG   a_text;	/* size of text segment */
    ULONG   a_data;	/* size of initialized data */
    ULONG   a_bss;	/* size of uninitialized data */
    ULONG   a_syms;	/* size of symbol table */
    ULONG   a_entry;	/* entry point */
    ULONG   a_trsize;	/* size of text relocation */
    ULONG   a_drsize;	/* size of data relocation */
};

#define	OMAGIC	0407		/* old impure format */
#define	NMAGIC	0410		/* read-only text */
#define	ZMAGIC	0413		/* demand load format */

/* machine types */
#define M_OLDSUN2	0	/* old sun-2 executable files */
#define M_68010		1	/* runs on either 68010 or 68020 */
#define M_68020		2	/* runs only on 68020 */
#define M_SPARC		3	/* runs only on SPARC */

struct reloc_info_68k
{
    long  r_address;	/* address which is relocated */
    unsigned int
    r_symbolnum:24,	/* local symbol ordinal */
    r_pcrel:1, 	        /* was relocated pc relative already */
    r_length:2,	        /* 0=byte, 1=word, 2=long */
    r_extern:1,	        /* does not include value of sym referenced */
    r_baserel:1,	/* linkage table relative */
    r_jmptable:1,	/* pc-relative to jump table */
    r_relative:1,	/* relative relocation */
	:1;
};

struct	nlist
{
    union
    {
	char	*n_name;	/* for use when in-core */
	long	n_strx;		/* index into file string table */
    } n_un;
    UBYTE	n_type;		/* type flag (N_TEXT,..)  */
    char	n_other;		/* unused */
    short	n_desc;			/* see <stab.h> */
    ULONG	n_value;	/* value of symbol (or sdb offset) */
};

/*
 * Simple values for n_type.
 */
#define	N_UNDF	0x0		/* undefined */
#define	N_ABS	0x2		/* absolute */
#define	N_TEXT	0x4		/* text */
#define	N_DATA	0x6		/* data */
#define	N_BSS	0x8		/* bss */
#define	N_COMM	0x12		/* common (internal to ld) */
#define	N_FN	0x1e		/* file name symbol */

#define	N_EXT	01		/* external bit, or'ed in */
#define	N_TYPE	0x1e		/* mask for all the type bits */

#define	N_STAB	0xe0		/* if any of these bits set, a dbx symbol */

/* Command line options */
#define OPT_VERBOSE    (1<<0)  /* Show 'everything' */
#define OPT_ANALYZE    (1<<1)  /* Don't output anything, just check */
#define OPT_DATA       (1<<2)  /* Create linkable data from raw file */
#define OPT_RAW        (1<<3)  /* Create EPROMable chunk from a.out */

   /* I don't like globals, but I don't like long arglists either */
  /***********/
 /* globals */
/***********/
struct exec header;
struct reloc_info_68k *relocs;
struct nlist *symboltable;
int relocnum,symbolnum;
char *chunk1;

int showreloc(struct reloc_info_68k *reloc,int curreloc,int coderelocs,long options)
{
    static const char *size[4]={"byte","word","long","unknown"};
    int i;
    long temp,bssdelta;
    ULONG offset,r_symbolnum;

    if(reloc->r_extern)
    {
	if(reloc->r_symbolnum >= symbolnum)
	{
	    printf("0x%08lx (%s) External (unavailable symbol)\n",reloc->r_address, size[reloc->r_length]);
	    return 1;
	}
	if((symboltable[reloc->r_symbolnum].n_type == N_EXT) &&
	   symboltable[reloc->r_symbolnum].n_value)
	{
	    if(symboltable[reloc->r_symbolnum].n_un.n_name)
	    {
		if((options & OPT_VERBOSE))
		    printf("External %s",symboltable[reloc->r_symbolnum].n_un.n_name);
	    }
	    else
	    {
		if((options & OPT_VERBOSE))
		    printf("External (unknown)");
	    }

	    r_symbolnum = reloc->r_symbolnum;
	    if((options & OPT_VERBOSE))
		printf(" (global BSS %ld bytes) Allocated...\n",
		       (long)symboltable[r_symbolnum].n_value);
	    /* Remember the size change */
	    bssdelta = (long)symboltable[r_symbolnum].n_value;
		
	    /* 'relocate' all references to this symbol */
	    for(i=0;i<relocnum;i++)
	    {
		if(relocs[i].r_extern &&
		   relocs[i].r_symbolnum == r_symbolnum)
		{
		    /* The reloc entry we were originally processing
		       will also satisfy the expression. However,
		       reloc->r_symbolnum will then change. That is
		       why we copied it to r_symbolnum. */
		    
		    relocs[i].r_extern = 0;
		    relocs[i].r_symbolnum = N_BSS;
		    if(i < coderelocs)
		    {
			/* Code-relative address */
			offset = (ULONG)(chunk1 + relocs[i].r_address);
		    }
		    else
		    {
			/* Data-relative address */
			offset = (ULONG)(chunk1 + header.a_text +
					 relocs[i].r_address);
		    }
		    switch(relocs[i].r_length)
		    {
		    case 2:
			/* SUN/SPARC alignment restrictions !! */
			temp = ((long)((*(UWORD *)offset))<<16) +
			    (long)(*(UWORD *)(offset+2));
			temp += header.a_bss;
			*(UWORD *)offset = (temp>>16);
			*(UWORD *)(offset+2) = (UWORD)temp;
			if(temp != (((long)(*(UWORD *)offset))<<16) +
			   (long)(*(UWORD *)(offset+2)))
			{
			    printf("ERROR: written and read values differ (%ld and %ld)\n", temp, (((long)(*(UWORD *)offset))<<16) + (long)(*(UWORD *)(offset+2)));
			}
			break;
		    case 1:
			*(UWORD *)offset += header.a_bss;
			break;
		    case 0:
			*(UBYTE *)offset += header.a_bss;
			break;
		    default:
			printf("\nUnknown reloc size: 3\n");
			return 1;
			break;
		    }
		    if((options & OPT_VERBOSE))
		    {
			printf("\t");
			showreloc(&relocs[i],i,coderelocs,options);
		    }
		}
	    }
	    
	    /* Increase the size of the BSS section */
	    header.a_bss += bssdelta;
	    
	    return 0;
	}
	else
	{
	    printf("0x%08lx (%s) External ",reloc->r_address, size[reloc->r_length]);
	    if(symboltable[reloc->r_symbolnum].n_un.n_name)
	    {
		printf("%s",symboltable[reloc->r_symbolnum].n_un.n_name);
	    }
	    else
	    {
		printf("(unavailable name)");
	    }
	    if(reloc->r_pcrel)
	    {
		printf(" (PC-relative)");
	    }
	    printf("\n");
	}
	return 1;
    }
    else
    {
	if(reloc->r_symbolnum & N_STAB)
	{
	}
	else
	{
	    if((options & OPT_VERBOSE))
	    {
		printf("0x%08lx (%s) ",reloc->r_address, size[reloc->r_length]);
		if(reloc->r_pcrel)
		{
		    printf("(PC-relative) ");
		}
		switch(reloc->r_symbolnum & ~N_EXT)
		{
		case N_ABS:
		    printf("Absolute      ");
		    break;
		case N_TEXT:
		    printf("Text-relative ");
		    break;
		case N_DATA:
		    printf("Data-relative ");
		    break;
		case N_BSS:
		    printf("BSS-relative  ");
		    break;
		case N_COMM:
		    printf("Common        ");
		    break;
		case N_FN:
		    printf("Filename      ");
		    break;
		default:
		    printf("unknown       ");
		    break;
		}
		if(curreloc < coderelocs)
		{
		    /* Code-relative address */
		    offset = (ULONG)(chunk1 + reloc->r_address);
		}
		else
		{
		    /* Data-relative address */
		    offset = (ULONG)
			(chunk1 + header.a_text + reloc->r_address);
		}
		switch(reloc->r_length)
		{
		case 2:
		    /* SUN/SPARC alignment restrictions !! */
		    temp = (((long)*(UWORD *)offset)<<16) +
			(long)(*(UWORD *)(offset+2));
		    printf("0x%08lx\n",temp);
		    break;
		case 1:
		    printf("0x%04lx\n",(ULONG)(*(UWORD *)offset));
		    break;
		case 0:
		    printf("0x%02lx\n",(ULONG)(*(UBYTE *)offset));
		    break;
		default:
		    printf("\nUnknown reloc size: 3\n");
		    break;
		}
	    }
	}
    }
    return 0;
}


void main(int argc, char *argv[])
{
    FILE *fp;
    char temp[80],*infile=NULL,*outfile=NULL,*symbolname=NULL;
    int l,i,n,possible;
    long extcode,extdata,options = 0, error = 0;

    long chunk1size,stringsize;
    char *stringtable;

    for(n=1; n<argc; n++)
    {
	if(argv[n][0]=='-')
	{
	    i=1;
	    while(argv[n][i])
	    {
		switch(argv[n][i++])
		{
		case 's':
		    if(symbolname)
		    {
			fprintf(stderr,"Too many symbol names!\n");
			error = 10;
		    }
		    else
		    {
			if(argv[n][i])
			{
			    symbolname = &argv[n][i];
			    i = strlen(argv[n]);
			}
			else
			{
			    if(argv[n+1])
			    {
				symbolname = argv[n+1];
				n++;
				i = strlen(argv[n]);
			    }
			    else
			    {
				fprintf(stderr,"Symbol name missing!\n");
				error = 10;
			    }
			}
		    }
		    break;
		case 'r':
		    options |= OPT_RAW;
		    break;
		case 'd':
		    options |= OPT_DATA;
		    break;
		case 'v':
		    options |= OPT_VERBOSE;
		    break;
		case 'a':
		    options |= OPT_ANALYZE;
		    break;
		case 'o':
		    if(outfile)
		    {
			fprintf(stderr,"Too many output filenames!\n");
			error = 10;
		    }
		    else
		    {
			if(argv[n][i])
			{
			    outfile = &argv[n][i];
			    i = strlen(argv[n]);
			}
			else
			{
			    if(argv[n+1])
			    {
				outfile = argv[n+1];
				n++;
				i = strlen(argv[n]);
			    }
			    else
			    {
				fprintf(stderr,"Output filename missing!\n");
				error = 10;
			    }
			}
		    }
		    break;
		default:
		    fprintf(stderr,"Unknown option!\n");
		    error = 5;
		    break;
		}
		if(error)
		    break;
	    }
	}
	else
	{
	    if(infile)
	    {
		fprintf(stderr,"Too many input filenames!\n");
		error = 10;
	    }
	    else
	    {
		infile = argv[n];
	    }
	}
	if(error)
	    break;
    }
    if(error || !infile)
    {
	fprintf(stderr,"Usage: %s [-vadro<outfile> -s<symname>] <infile>\n",argv[0]);
	fprintf(stderr,"       -a  Analyze only\n");
	fprintf(stderr,"       -v  Verbose - show file and relocation info\n");
	fprintf(stderr,"       -r  Generate EPROMmable data - text and data only\n");
	fprintf(stderr,"           (<file>.raw = byte-format, <file>.odd = odd bytes, <file>.even = even bytes)\n");
	fprintf(stderr,"       -d  Generate linkable data from a file\n");
	fprintf(stderr,"           (Text section, symbols \"__<input file>\", \"__<input file>_end\")\n");
	fprintf(stderr,"       -s  Change the symbol name for linkable data (only used with -d)\n");
	exit(error);
    }

    if((options & OPT_RAW))
    {
	if((options & (OPT_ANALYZE | OPT_DATA)))
	{
	    fprintf(stderr,"Warning: -r and -d and/or -a specified, raw data generated!\n");

	}
	if((fp = fopen(infile, "rb")))
	{
	    if(fread((char *)&header, 1, sizeof(struct exec), fp))
	    {
		if( (header.a_magic != OMAGIC && header.a_magic != NMAGIC)
		   || header.a_machtype==M_SPARC)
		{
		    printf("Wrong executable format!\n");
		    fclose(fp);
		    exit(10);
		}
		if(header.a_trsize || header.a_drsize)
		{
		    printf("Link the executable to a fixed address first!\n");
		    fclose(fp);
		    exit(10);
		}
		chunk1size = header.a_text + header.a_data;
		if((chunk1 = (char *)malloc(chunk1size)))
		{
		    char *buffer = calloc((chunk1size+1)/2, 1);
		    int oddsize = chunk1size/2, evensize=chunk1size-oddsize;

		    if(!buffer)
		    {
			fprintf(stderr, "Memory allocation error!\n");
			free(chunk1);
			fclose(fp);
			exit(10);
		    }

		    if(fread(chunk1, 1, chunk1size, fp) != chunk1size)
			fprintf(stderr, "Unexpected EOF!\n");

		    fclose(fp);
		    if(outfile)
		    {
			sprintf(temp, "%s.raw", outfile);
		    }
		    else
		    {
			sprintf(temp, "%s.raw", infile);
		    }
		    if((fp = fopen(temp, "wb")))
		    {
			fwrite(chunk1, 1, chunk1size, fp);
			fclose(fp);
		    }
		    else
		    {
			printf("Could not open \"%s\" for writing.\n", temp);
			error = 10;
		    }
		    if(outfile)
		    {
			sprintf(temp, "%s.even", outfile);
		    }
		    else
		    {
			sprintf(temp, "%s.even", infile);
		    }
		    if((fp = fopen(temp, "wb")))
		    {
			/* Write even bytes */
			for(i=0;i<evensize;i++)
			{
			    buffer[i] = chunk1[2*i];
			}
			fwrite(buffer, 1, evensize, fp);
			fclose(fp);
		    }
		    else
		    {
			printf("Could not open \"%s\" for writing.\n", temp);
			error = 10;
		    }
		    if(outfile)
		    {
			sprintf(temp, "%s.odd", outfile);
		    }
		    else
		    {
			sprintf(temp, "%s.odd", infile);
		    }
		    if((fp = fopen(temp, "wb")))
		    {
			/* Write odd bytes */
			for(i=0;i<oddsize;i++)
			{
			    buffer[i] = chunk1[2*i+1];
			}
			fwrite(buffer, 1, oddsize, fp);
			fclose(fp);
		    }
		    else
		    {
			printf("Could not open \"%s\" for writing.\n", temp);
			error = 10;
		    }
		    free(buffer);
		    free(chunk1);
		    printf("Start address 0x%lx\n", header.a_entry);
		}
		else
		{
		    printf("Memory allocation error!\n");
		    fclose(fp);
		    error = 10;
		}
	    }
	    else
	    {
		printf("Fatal error reading \"%s\".\n", infile);
		fclose(fp);
		error = 10;
	    }
	}
	else
	{
	    printf("Could not open \"%s\" for reading.\n", infile);
	    error = 10;
	}
	exit(error);
    }

    if((options & OPT_DATA))
    {
	if((options & OPT_ANALYZE))
	{
	    fprintf(stderr,"Warning: both -d and -a specified, -a discarded!\n");
	}
	if((fp = fopen(infile, "rb")))
	{
	    fseek(fp, 0, 2); /* Wind to the end of file */
	    chunk1size = ftell( fp );
	    if((chunk1 = (char *)calloc((chunk1size+3)&~3, 1)))
	    {
		fseek(fp, 0, 0); /* Rewind to the start of the file */
		if(fread(chunk1, 1, chunk1size, fp))
		{
		    fclose(fp);
		    if(outfile)
		    {
			strcpy(temp, outfile);
		    }
		    else
		    {
			sprintf(temp, "%s.o", infile);
		    }
		    if((fp = fopen(temp, "wb")))
		    {	
			struct nlist symbol;

			header.a_toolversion = 0;
			header.a_machtype = M_68010;
			header.a_magic = OMAGIC;
			
			header.a_text = (chunk1size+3)&~3;
			header.a_data = 0;
			header.a_bss = 0;
			header.a_syms = 2*sizeof(struct nlist);
			header.a_entry = 0;
			header.a_trsize = 0;
			header.a_drsize = 0;
			
			symbol.n_un.n_strx = 4; /* String-section offset 0 */
			symbol.n_type = N_TEXT | N_EXT;
			symbol.n_other = 0;
			symbol.n_desc = 0;
			symbol.n_value = 0; /* offset is 0 */

			if(symbolname)
			    strcpy(temp, symbolname);
			else
			    sprintf(temp, "__%s", infile);
			stringsize = strlen(temp)+5;
			/* includes \0 and the variable */

			fwrite((char *)&header, sizeof(struct exec), 1, fp);
			fwrite(chunk1, (chunk1size+3)&~3, 1, fp);
			fwrite((char *)&symbol, sizeof(struct nlist), 1, fp);

			symbol.n_un.n_strx = stringsize;
			symbol.n_type = N_TEXT | N_EXT;
			symbol.n_other = 0;
			symbol.n_desc = 0;
			symbol.n_value = chunk1size;

			fwrite((char *)&symbol, sizeof(struct nlist), 1, fp);
			if(symbolname)
			    sprintf(temp+strlen(temp)+1, "%s_end", symbolname);
			else
			    sprintf(temp+strlen(temp)+1, "__%s_end", infile);
			stringsize += strlen(temp+strlen(temp)+1)+1;

			fwrite((char *)&stringsize, sizeof(long), 1,fp);
			fwrite(temp, 1, stringsize-4, fp);
			fclose(fp);
		    }
		    else
		    {
			printf("Could not open \"%s\" for writing.\n", temp);
			error = 10;
		    }
		}
		else
		{
		    printf("Fatal error reading \"%s\".\n", infile);
		    fclose(fp);
		    error = 10;
		}
		free(chunk1);
	    }
	    else
	    {
		printf("Memory allocation error!\n");
		fclose(fp);
		error = 10;
	    }
	}
	else
	{
	    printf("Could not open \"%s\" for reading.\n", infile);
	    error = 10;
	}
	exit(error);
    }

    if(outfile && (options & OPT_ANALYZE))
    {
	fprintf(stderr,"Warning: both -o and -a specified, output file will not be generated!\n");
    }

    if((fp=fopen(infile,"rb")))
    {
	possible = 1;
	if(fread((char *)&header, sizeof(struct exec), 1, fp))
	{
	    if(header.a_magic!=OMAGIC && header.a_magic!=NMAGIC && header.a_magic!=ZMAGIC)
	    {
		printf("\"%s\": Not a.out format!\n", infile);
		fclose(fp);
		exit(10);
	    }
	    if((options & OPT_VERBOSE))
	    {
		strcpy(temp, "::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::");
		if(strlen(infile)<72)
		    temp[strlen(infile)+6]= '\0';
		printf("%s\n:: %s ::\n%s\n", temp, infile, temp);

		if(header.a_toolversion & 0x80)
		    printf("Dynamic\n");
		else
		    printf("Static\n");
		printf("Toolversion : %ld (", (long)(header.a_toolversion & 0x7f));
		switch(header.a_toolversion & 0x7f)
		{
		case 0:
		    printf("Assembler");
		    break;
		case 1:
		    printf("Linker");
		    break;
		case 68:
		    printf("Xlink");
		    break;
		default:
		    printf("unknown");
		    break;
		}
		printf(")\nMachine type: %ld (", (long)header.a_machtype);
		switch(header.a_machtype)
		{
		case M_68010:
		    printf("68010");
		    break;
		case M_68020:
		    printf("68020");
		    break;
		case M_SPARC:
		    printf("SPARC");
		    break;
		case 0:
		    printf("68010");
		    break;
		default:
		    printf("unknown");
		    break;
		}
		printf(")\nMagic number: 0x%04lx (", (long)header.a_magic);
		switch(header.a_magic)
		{
		case OMAGIC:
		    printf("OMAGIC");
		    break;
		case NMAGIC:
		    printf("NMAGIC");
		    break;
		case ZMAGIC:
		    printf("ZMAGIC");
		    break;
		default:
		    printf("unknown");
		    break;
		}
		printf(")\nCode size   : 0x%lx\n", header.a_text);
		printf("Data size   : 0x%lx\n", header.a_data);
		printf("BSS size    : 0x%lx\n", header.a_bss);
		printf("Symbol size : 0x%lx\n", header.a_syms);
		printf("Entry point : 0x%lx\n", header.a_entry);
		printf("Code reloc size: 0x%lx\n", header.a_trsize);
		printf("Data reloc size: 0x%lx\n", header.a_drsize);
	    }
	    if( (header.a_magic == OMAGIC || header.a_magic == NMAGIC)
	       && header.a_machtype!=M_SPARC)
	    {
		/* Long-Align reloc infos (and symbols)!! */
		chunk1size = ((header.a_text + header.a_data + 3) & ~3) +
		    header.a_trsize + header.a_drsize + header.a_syms;

		relocnum = (header.a_trsize+header.a_drsize)/
		    sizeof(struct reloc_info_68k);
		extcode = 0;
		extdata = 0;
		symbolnum = header.a_syms/sizeof(struct nlist);
		
		if((chunk1 = (char *)calloc(chunk1size,1)))
		{
		    fread(chunk1, 1, header.a_text+header.a_data, fp);
		    relocs = (struct reloc_info_68k *)
			(chunk1 + ((header.a_text + header.a_data + 3) & ~3));
		    if(fread((char *)relocs, 1, header.a_trsize+ header.a_drsize+header.a_syms, fp))
		    {
			symboltable = (struct nlist *)
			    ((char *)relocs +
			     header.a_trsize + header.a_drsize);
			
			if(fread((char *)&stringsize, 1, 4, fp))
			{
			    if(stringsize>=4)
				stringsize -= 4;
			}
			else
			{
			    stringsize = 0;
			}
			stringtable = NULL;
			if(stringsize)
			{
			    if((stringtable=(char *)calloc(stringsize,1)))
			    {
				fread(stringtable, 1, stringsize, fp);
			    }
			    else
			    {
				printf("calloc failed for string data.\n");
			    }
			}
			fclose(fp); /* Done with the file */
			
			/* set the name pointers */
			if(stringtable)
			{
			    for(i=0;i<symbolnum;i++)
			    {
				symboltable[i].n_un.n_name =
				    stringtable +
					symboltable[i].n_un.n_strx - 4;
				if(!strcmp("_end",
					   symboltable[i].n_un.n_name))
				{
				    possible = 0;
				}
			    }
			}
			else
			{
			    for(i=0;i<symbolnum;i++)
			    {
				symboltable[i].n_un.n_name = NULL;
			    }
			}
			
			if(header.a_toolversion != 68)
			{
			    /* 'Relocate' data & bss reference base
			       to Data and BSS segment start addresses
			       instead of Text segment start */
			    l = header.a_trsize/
				sizeof(struct reloc_info_68k);
			    for(i=0; i<relocnum; i++)
			    {
				ULONG temp,offset;

				temp = 0;
				switch(relocs[i].r_symbolnum & ~N_EXT)
				{
				case N_DATA:
				    temp = header.a_text;
				    break;
				case N_BSS:
				    temp = header.a_text + header.a_data;
				    break;
				}
				if(i<l)
				{
				    /* Code-relative address */
				    offset = (ULONG)
					(chunk1 + relocs[i].r_address);
				}
				else
				{
				    /* Data-relative address */
				    offset = (ULONG)
					(chunk1 + header.a_text +
					 relocs[i].r_address);
				}
				switch(relocs[i].r_length)
				{
				case 2:
				    /* SPARC alignment restrictions!! */
				    temp=((long)((*(UWORD *)offset))<<16)+
					(long)(*(UWORD *)(offset+2))-temp;
				    
				    *(UWORD *)offset = (temp>>16);
				    *(UWORD *)(offset+2) = (UWORD)temp;
				    if(temp != 
				       (((long)(*(UWORD *)offset))<<16) +
				       (long)(*(UWORD *)(offset+2)))
				    {
					printf("ERROR: written and read values differ (%ld and %ld)\n", temp, (((long)(*(UWORD *)offset))<<16) + (long)(*(UWORD *)(offset+2)));
				    }
				    break;
				case 1:
				    *(UWORD *)offset -= temp;
				    break;
				case 0:
				    *(UBYTE *)offset -= temp;
				    break;
				default:
				    printf("\nUnknown reloc size: 3\n");
				    break;
				}
			    }
			}

			/* Search for external references etc. */
			l = header.a_trsize/sizeof(struct reloc_info_68k);
			if(relocnum)
			{
			    if((options & OPT_VERBOSE))
				printf("Code relocations:\n");
			    for(i=0; i<relocnum; i++)
			    {
				if((options & OPT_VERBOSE))
				    if(i==l)
					printf("Data relocations:\n");
				
				if(i<l)
				{
				    extcode += showreloc(&relocs[i],i,l,options);
				}
				else
				{
				    extdata += showreloc(&relocs[i],i,l,options);
				}
			    }
			    if(extcode)
				printf("%ld unresolved Code references!\n",extcode);
			    if(extdata)
				printf("%ld unresolved Data references!\n",extdata);
			}
			/* Output relocatable format */
			
			if(header.a_toolversion != 68 &&
			   !extcode && !extdata && possible &&
			   !(options & OPT_ANALYZE))
			{
			    printf("Outputting .b68k-format\n");
			    if(!outfile)
			    {
				sprintf(temp, "%s.b68k", infile);
				outfile = temp;
			    }
			    if((fp=fopen(outfile,"wb")))
			    {
				header.a_toolversion = 68;
				header.a_syms = 0;
				header.a_entry = 0;
				fwrite((char *)&header, sizeof(struct exec), 1, fp);
				fwrite(chunk1, 1, header.a_text+header.a_data+header.a_trsize+header.a_drsize,fp);
				
				fclose(fp);
			    }
			    else
			    {
				printf("Could not open \"%s\" for writing\n",outfile);
			    }
			}
			
			if(stringtable)
			    free(stringtable);
		    }
		    else
		    {
			fclose(fp);
			printf("Unexpected EOF in \"%s\" (short file)\n",infile);
		    }
		    free(chunk1);
		}
		else
		{
		    fclose(fp);
		}
		if(possible)
		{
		    if(!extcode && !extdata)
		    {
			if(!relocnum)
			{
			    printf("Position-Independent Code\n");
			}
			else
			{
			    printf("Relocation possible\n");
			}
		    }
		    else
		    {
			if(!symbolnum)
			{
			    printf("Unusable file (Unresolved symbols and no symbol information)\n");
			}
			else
			{
			    printf("Relocation possible after external references are resolved\n");
			}
		    }
		}
		else
		{
		    printf("Relocation NOT possible! (unless this is PIC)\nLinked to address 0x%lx\n",header.a_entry);
		}
	    }
	    else
	    {
		fclose(fp);
		printf("Executable format not handled!\n");
	    }
	}
	else
	{
	    fclose(fp);
	    printf("\"%s\": does not include a header\n",infile);
	}
    }
    else
    {
	printf("Can't open \"%s\"\n",infile);
    }
    exit(error);
}
